#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	EventLaneViewer.py
#	
#	This is a gui class that acts as the container for all the 
#	EventViewers belonging to a single instrument. Graphically,
#	this class makes up the timeline portion of the instrument,
#	on which the events can be placed and moved.
#
#-------------------------------------------------------------------------------

import gtk
from EventViewer import *
import os.path
import gettext
import PlatformUtils
import urllib # To decode URI's
import Globals # To get projectfolder
import ui.EventLaneHSeparator as EventLaneHSeparator
_ = gettext.gettext

#=========================================================================

class EventLaneViewer(gtk.EventBox):
	"""
	This class is a container for all the individual EventViewers
	for a single Instrument.
	"""
	
	""" Number only to be used inside Jokosher """
	URI_DRAG_TYPE = 84
	
	""" Custom numbers for use while dragging text in Jokosher """
	DRAG_TARGETS = [ ( "text/uri-list", 	# Accept uri-lists
						0,					# From everywhere
						URI_DRAG_TYPE ),		# Use the custom number
						("text/plain", 0, URI_DRAG_TYPE) # so drags from Firefox work
						]
	
	#_____________________________________________________________________

	def __init__(self, project, instrument, instrumentviewer, mainview, small = False):
		"""
		Creates a new instance of EventLaneViewer.
		
		Parameters:
			project -- the currently active Project.
			instrument -- the Instrument that the Event lane belongs to.
			instrumentviewer -- the InstrumentViewer holding the Event lane.
			mainview -- the MainApp Jokosher window.
			small -- set to True if we want small edit views (i.e. for mixing view).
		"""
		gtk.EventBox.__init__(self)

		self.small = small
		self.instrumentviewer = instrumentviewer
		self.mainview = mainview
		self.vbox = gtk.VBox()
		self.fixed = gtk.Fixed()

		self.separator = EventLaneHSeparator.EventLaneHSeparator(project, project.transport)
		self.vbox.pack_start(self.fixed, True, True)
		self.vbox.pack_start(self.separator, False, False)

		self.vbox.show_all()
		self.add(self.vbox)
		self.show_all()
			
		self.project = project
		self.instrument = instrument
		self.project.transport.connect("position", self.OnTransportPosition)
		self.project.connect("view-start", self.OnProjectViewChange)
		self.project.connect("zoom", self.OnProjectViewChange)
		self.instrument.connect("event::added", self.OnEventAdded)
		self.instrument.connect("event::removed", self.OnEventRemoved)
		
		# This defines where the blue cursor indicator should be drawn (in pixels)
		self.highlightCursor = None
		
		# True if the popup menu is visible
		self.popupIsActive = False
		
		#The position where the last mouse click was
		self.mouseDownPos = [0,0]
		
		#the list of all the EventViewer widgets
		self.eventViewerList = []
		
		self.set_events(	gtk.gdk.POINTER_MOTION_MASK |
							gtk.gdk.BUTTON_RELEASE_MASK |
							gtk.gdk.BUTTON_PRESS_MASK )
		
		self.connect("button-press-event", self.OnMouseDown)
		self.connect("motion_notify_event", self.OnMouseMove)
		self.connect("leave_notify_event", self.OnMouseLeave)
		self.fixed.drag_dest_set(	gtk.DEST_DEFAULT_DROP,
									self.DRAG_TARGETS, 
									gtk.gdk.ACTION_COPY)
		self.fixed.connect("drag_data_received", self.OnDragDataReceived)
		self.fixed.connect("drag_motion", self.OnDragMotion)
		self.fixed.connect("drag_leave", self.OnDragLeave)
		self.fixed.connect("expose-event", self.OnDraw)
		
		#create the context menu
		self.contextMenu = gtk.Menu()
		
		audioimg = gtk.Image()
		if self.mainview.audioFilePixbuf:
			size = gtk.icon_size_lookup(gtk.ICON_SIZE_MENU)
			pixbuf = self.mainview.audioFilePixbuf.scale_simple(size[0], size[1], gtk.gdk.INTERP_BILINEAR)
			audioimg.set_from_pixbuf(pixbuf)
		
		menuItem = gtk.ImageMenuItem(_("_Add Audio File..."), True)
		menuItem.set_image(audioimg)
		menuItem.connect("activate", self.CreateEventFromFile)
		self.contextMenu.append(menuItem)
		
		menuItem = gtk.SeparatorMenuItem()
		self.contextMenu.append(menuItem)
		
		self.pasteContextMenuItem = gtk.ImageMenuItem(gtk.STOCK_PASTE)
		self.pasteContextMenuItem.connect("activate", self.OnPaste)
		self.contextMenu.append(self.pasteContextMenuItem)
		
		menuItem = gtk.ImageMenuItem(gtk.STOCK_DELETE)
		menuItem.connect("activate", self.OnDelete)
		self.contextMenu.append(menuItem)
		
		self.messageID = None
		
		for event in self.instrument.events:
			self.OnEventAdded(self.instrument, event)
		
	#_____________________________________________________________________
		
	def OnDraw(self, widget, event):
		"""
		Called everytime the window is drawn.
		Handles the drawing of the lane edges and vertical line cursors.
		
		Parameters:
			widget -- GTK widget to be repainted.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""

		wnd = widget.window
		gc = wnd.new_gc()
		
		transport = self.project.transport
		
		# Draw play cursor position
		col = gc.get_colormap().alloc_color("#FF0000")
		gc.set_foreground(col)
		x = transport.GetPixelPosition()
		wnd.draw_line(gc, x, 0, x, self.allocation.height)
		
	#_____________________________________________________________________
		
	def UpdatePosition(self, eventViewer):
		"""
		Moves the given EventViewer widget to the appropriate position.
		
		Parameters:
			eventViewer -- the widget that has needs to be moved to a new position.
		"""
		if eventViewer in self.fixed.get_children():
			x = int(round((eventViewer.event.start - self.project.viewStart) * self.project.viewScale))
			self.fixed.move( eventViewer, x, 0 )
			
			self.queue_draw()
			
	#_____________________________________________________________________
	
	def Destroy(self):
		"""
		Called when the EventLaneViewer gets destroyed.
		It also destroys any child widget and disconnects itself from any
		gobject signals.
		"""
		self.project.transport.disconnect_by_func(self.OnTransportPosition)
		self.project.disconnect_by_func(self.OnProjectViewChange)
		self.instrument.disconnect_by_func(self.OnEventAdded)
		self.instrument.disconnect_by_func(self.OnEventRemoved)
		
		for widget in self.fixed.get_children():
			#Check that it is EventViewer (could be a button drawer)
			if type(widget) == EventViewer:
				widget.Destroy()
		
		self.destroy()
	
	#_____________________________________________________________________

	def OnMouseDown(self, widget, mouse):
		"""
		Called when the user pressed a mouse button.
		If it's a right-click, creates a context menu on the fly for importing,
		pasting and deleting Events.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			mouse -- GTK mouse event that fired this method call.
		"""
		if self.project.GetIsRecording():
			return True

		self.mouseDownPos = [mouse.x, mouse.y]
		
		if mouse.type == gtk.gdk._2BUTTON_PRESS:
			self.CreateEventFromFile()
			
		if self.popupIsActive:
			self.OnMenuDone()
		# Create context menu on RMB 
		elif mouse.button == 3:
			self.pasteContextMenuItem.set_sensitive( bool(self.project.clipboardList) )
		
			self.highlightCursor = mouse.x
			self.popupIsActive = True
			
			self.contextMenu.connect("selection-done", self.OnMenuDone)
			self.contextMenu.show_all()
			self.contextMenu.popup(None, None, None, mouse.button, mouse.time)
		
		if 'GDK_CONTROL_MASK' in mouse.state.value_names:
			self.instrument.SetSelected(True)
		else:
			self.project.ClearEventSelections()
			self.project.SelectInstrument(self.instrument)
			
		return True
			
	#_____________________________________________________________________
	
	def OnMenuDone(self, widget=None):
		"""
		Hides the right-click context menu after the user has selected one
		of its options or clicked elsewhere.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.popupIsActive = False
		self.highlightCursor = None
	
	#_____________________________________________________________________

	def OnMouseMove(self, widget, mouse):
		"""
		Display a message in the StatusBar when the mouse hovers over the
		EventLaneViewer.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			mouse -- GTK mouse event that fired this method call.
		"""
		if self.project.GetIsRecording():
			return True
		
		# display status bar message if has not already been displayed
		if not self.messageID: 
			self.messageID = self.mainview.SetStatusBar(_("<b>Right-click</b> for more options."))
		# TODO - we need to add code here to snap to beat/bar etc.
		self.highlightCursor = mouse.x
		self.queue_draw()
		
	#_____________________________________________________________________
		
	def OnMouseLeave(self, widget, mouse):
		"""
		Clears the StatusBar message when the mouse moves out of the
		EventLaneViewer area.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			mouse -- GTK mouse event that fired this method call.
		"""
		if self.messageID:   #clear status bar if not already clear
			self.mainview.ClearStatusBar(self.messageID)
			self.messageID = None
		if not self.popupIsActive:
			self.highlightCursor = None
		self.queue_draw()

	#_____________________________________________________________________
	
	def CreateEventFromFile(self, event=None):
		"""
		Called when "Import Audio File..." is selected from the right-click context menu.
		Opens up a file chooser dialog to import an Event.
		
		Parameters:
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		filenames, copyfile = self.mainview.ShowImportFileChooser()
		#filename will be None is the user cancelled the dialog
		if filenames:
			start = 0
			if event:
				#if we we're called from a mouse click, use the mouse position as the start
				start = (self.mouseDownPos[0]/self.project.viewScale) + self.project.viewStart
	
			uris = [PlatformUtils.pathname2url(filename) for filename in filenames]
			self.instrument.AddEventsFromList(start, uris)

	#_____________________________________________________________________
	
	def OnPaste(self, widget):
		"""
		Called when "Paste" is selected from the context menu.
		Adds the selected Event to the clipboard.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if not self.project.clipboardList:
			return
		
		for event in self.project.clipboardList:
			start = (self.mouseDownPos[0]/self.project.viewScale) + self.project.viewStart
			self.instrument.addEventFromEvent(start, event)
		
	#_____________________________________________________________________
	
	def OnDelete(self, event=None):
		"""
		Called when "Delete" is selected from context menu.
		Deletes the selected Instrument from the Project.
		
		Considerations:
			This delete is called when right-clicking an EMPTY section
			of the EventLaneViewer. For right-clicking over a selected
			Event see OnDelete in EventViewer.
			
		Parameters:
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.project.DeleteInstrument(self.instrument.id)
	
	#_____________________________________________________________________
	
	def OnProjectViewChange(self, project):
		"""
		Callback function for when the project view changes,
		and the "view-start" or the "zoom" signal is send, and we
		need to update.
		
		Parameters:
			project -- The project instance that send the signal.
		"""
		for event in self.eventViewerList:
			self.UpdatePosition(event)
		
	#_____________________________________________________________________
	
	def OnEventAdded(self, instrument, event):
		"""
		Callback for when an event is added to our instrument.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
			event -- the event instance that was added.
		"""
		x = int(round((event.start - self.project.viewStart) * self.project.viewScale))
		child = EventViewer(self, self.project, event, self.allocation.height, self.mainview, self.small)
		self.fixed.put(child, x, 0)
		child.show()
		self.eventViewerList.append(child)
	
	#_____________________________________________________________________
	
	def OnEventRemoved(self, instrument, event):
		"""
		Callback for when an event is removed from our instrument.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
			event -- the event instance that was removed.
		"""
		for widget in self.eventViewerList:
			if widget.event is event:
				self.fixed.remove(widget)
				# remove the event's drawer if it's showing
				if widget.drawer.parent == self.fixed:
					self.fixed.remove(widget.drawer)
				#destroy the object
				widget.Destroy()
				self.eventViewerList.remove(widget)
				break
	
	#_____________________________________________________________________
	
	def OnTransportPosition(self, transportManager, extraString):
		"""
		Callback for signal when the transport position changes.
		Here we just redraw the playhead.
		
		Parameters:
			transportManager -- the TransportManager instance that send the signal.
			extraString -- a string specifying the extra action details. i.e. "stop-action"
					means that the position changed because the user hit stop.
		"""
		prev_pos = self.project.transport.GetPreviousPixelPosition()
		new_pos = self.project.transport.GetPixelPosition()
		self.queue_draw_area(prev_pos - 1, 0, 3, self.allocation.height)
		self.queue_draw_area(new_pos - 1, 0, 3, self.allocation.height)
	
	#_____________________________________________________________________
	
	def OnDragDataReceived(self, widget, context, x, y, selection, targetType, time):
		"""
		Called when the user releases MOUSE1, finishing a drag and drop
		procedure.
		Adds an Event for each "file://"-uri in the uri-list to the Instrument, 
		one after the other. The files will be copied to the Project's audio directory.
			
		Parameters:
			widget -- InstrumentViewer being dragged.
			context -- reserved for GTK callbacks, don't use it explicitly.
			x -- point in the X axis the dragged object was dropped.
			y -- point in the Y axis the dragged object was dropped..
			selection -- selected object area that was dragged.
			targetType -- mimetype of the dragged object.
			time -- reserved for GTK callbacks, don't use it explicitly.
			
		Returns:
			True -- continue GTK signal propagation. *CHECK*
		"""
		start = (x/self.project.viewScale) + self.project.viewStart
		# Splitlines to separate the uri's, unquote to decode the uri-encoding ('%20' -> ' ')
		uris = [urllib.unquote(uri) for uri in selection.data.splitlines()]
		self.instrument.AddEventsFromList(start, uris, True)
		
		context.finish(True, False, time)
		return True
	
	#_____________________________________________________________________
	
	def OnDragMotion(self, widget, context, x, y, time):
		"""
		Called each time the user moves the mouse while dragging.
		Draws a cursor on the EventLane while dragging something over it.
		
		Parameters:
			widget -- InstrumentViewer the mouse is hovering over.
			context -- cairo widget context.
			x -- reserved for GTK callbacks, don't use it explicitly.
			y -- reserved for GTK callbacks, don't use it explicitly.
			time -- reserved for GTK callbacks, don't use it explicitly.
		
		Returns:
			True -- continue GTK signal propagation. *CHECK*
		"""
		context.drag_status(gtk.gdk.ACTION_COPY, time)
		self.highlightCursor = x
		self.queue_draw()
		return True
	
	#_____________________________________________________________________
	
	def OnDragLeave(self, widget, drag_context, timestamp):
		"""
		Called when the user moves the cursor ouf of the EventLaneViewer
		while performing a drag and drop procedure.
		Hides the highlight cursor.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			drag_context -- reserved for GTK callbacks, don't use it explicitly.
			timestamp -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.highlightCursor = None
		self.queue_draw()
		
	#_____________________________________________________________________
	
	def ChangeSize(self, small):
		"""
		Changes the size of the event lane.
		
		Parameters:
			small - True if the event lane is to be small.
		"""
		self.small = small
		for widget in self.fixed.get_children():
			if type(widget) == EventViewer:
				widget.ChangeSize(small)

				if widget.drawer in self.fixed.get_children():
					widget.ShowDrawer()
	
	#____________________________________________________________________	
	
	def PutDrawer(self, drawer, xvalue=1):
		"""
		Places the drawer below in the event lane and makes it visible.
		
		Parameters:
			drawer -- the widget to show.
			xvalue -- the horizontal position of the drawer in pixels
		"""
		if self.small:
		    yvalue = 30
		else:
		    yvalue = 75

		if not drawer.parent:
			self.fixed.put(drawer, xvalue, yvalue)
		elif drawer.parent == self.fixed:
			self.fixed.move(drawer, xvalue, yvalue)
		
		drawer.show_all()
	
	#____________________________________________________________________
	
	def RemoveDrawer(self, drawer):
		"""
		Removes the drawer from below in the event. This function does
		nothing if the given drawer is not currenly shown.
		
		Parameters:
			drawer -- the widget to remove.
		"""
		if drawer.parent == self.fixed:
			self.fixed.remove(drawer)
			
	#____________________________________________________________________

#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	This class handles the saving, modifying, and deleting of mixdown profiles.
#
#-------------------------------------------------------------------------------

import Globals
import MixdownProfiles
import os

#=========================================================================

class MixdownProfileManager:
	"""
	This class handles the saving, loading and modification associated with MixdownProfiles.
	"""
	
	#_____________________________________________________________________

	def __init__(self, mixdownProfileDialog):
		"""
		Creates a new instance of MixdownProfileManager
		
		Parameters:
			parent -- reference to MixdownProfileDialog
		"""
		self.mixdownProfileDialog = mixdownProfileDialog
		self.profiles = MixdownProfiles.MixdownProfiles(self)
		self.profiles.connect("profile-update", self.MixdownProfileUpdate)
		self.profiles.connect("error-occurred", self.LoadingMixdownActionError)
	
	#_____________________________________________________________________
	
	def MixdownProfileUpdate(self, mixdownProfile, signalDetails):
		"""
		Called when a mixdown profile has been saved or deleted.
		Updates the profile combo model (self.parent.profileComboModel) accordingly.
		
		Parameters:
			mixdownProfile -- the MixdownProfiles instance that sent the signal
			signalDetails -- determines whether the profile was saved or deleted.
		"""
		self.mixdownProfileDialog.UpdateProfileModel(signalDetails)
		
	#_____________________________________________________________________
	
	def LoadingMixdownActionError(self, mixdownProfile, actionName, extensionName):
		"""
		Called when an error has occurred while loading MixdownActions.
		This method will call the ShowActionErrorDialog in MixdownProfileDialog,
		informing the user that a MixdownAction cannot be loaded.
		
		Parameters:
			actionName -- the name of the MixdownAction which cannot be loaded.
			extensionName -- the name of the extension that the MixdownAction can't be loaded from.
		"""
		self.mixdownProfileDialog.ShowActionErrorDialog(actionName, extensionName)
		
	#_____________________________________________________________________
	
	def GetMixdownProfileList(self):
		"""
		Called when the profile combo model (self.parent.profileComboModel) is populated.
		
		Returns:
			profilelist -- a list of profiles in the user's mixdown profiles directory.
		"""
		profilelist = []
		for files in os.listdir(Globals.MIXDOWN_PROFILES_PATH):
			if files.endswith(".profile"):
				filenames = files.split(".")[0]
				profilelist.append(filenames)
		return profilelist
		
	#_____________________________________________________________________

	def SaveMixdownProfile(self, profileName, actionList=None):
		"""
		Convenience method for saving a MixdownProfile to disk.
		
		Parameters:
			profileName -- The name of the profile to use in the profile filename.
			actionList -- list of MixdownActions to save.
		"""
		self.profiles.SaveProfile(profileName, actionList)
		
	#_____________________________________________________________________

	def DeleteMixdownProfile(self, profileName):
		"""
		Convenience method for deleting a MixdownProfile from disk.
		
		Parameters:
			profileName -- the name of the profile to be deleted
		"""
		self.profiles.DeleteProfile(profileName)

	#_____________________________________________________________________
	
	def ReturnAllActionsFromMixdownProfile(self, profileName):
		"""
		Convenience method for returning all actions in a MixdownProfile.
		
		Parameters:
			profileName -- the name of the MixdownProfile to load all MixdownActions from.
		"""
		actions = self.profiles.LoadMixdownActionsFromProfile(profileName)
		return actions
		
	#_____________________________________________________________________

#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	AddInstrumentDialog.py
#	
#	This module handles the dialog for adding instruments to a project. Don't
#	you just love instruments. We do. In fact, I suspect Laszlo has printed a
#	picture of the dialog box out and sleeps next to it.
#
#-------------------------------------------------------------------------------

import gtk.glade
import Globals
import textwrap
import gettext
_ = gettext.gettext


#=========================================================================

class AddInstrumentDialog:
	"""
	This class handles all of the processing associated with the Add Instrument dialog.
	"""
	#_____________________________________________________________________

	def __init__(self, project, parent, instr=None):
		"""
		Creates a new instance of AddInstrumentDialog.
		
		Parameters:
			project -- the Project to add instruments to.
			parent -- parent window of the AddInstrumentDialog (JokosherApp).
			instr -- if present, indicates the Instrument whose type wants to be changed.
		"""
		self.parent = parent
		self.project = project
		self.instr = instr

		self.res = gtk.glade.XML(Globals.GLADE_PATH, "AddInstrumentDialog")

		self.signals = {
			"on_OK_clicked" : self.OnOK,
			"on_Cancel_clicked" : self.OnCancel,
			"on_instrument_search_changed" : self.OnSearchChange,
			"on_AddInstrument_configure_event" : self.OnResize,
			"on_AddInstrument_destroy" : self.OnDestroy,
		}
		
		self.res.signal_autoconnect(self.signals)
		
		self.dlg = self.res.get_widget("AddInstrumentDialog")
		self.tree = self.res.get_widget("Instruments")
		self.search_entry = self.res.get_widget("instrument_search")
		self.search_entry.set_activates_default(True)
		self.okbutton = self.res.get_widget("okButton")
		self.okbutton.set_sensitive(False)
		self.okbutton.set_flags(gtk.CAN_DEFAULT)
		self.okbutton.grab_default()
		
		if self.instr: 
			self.dlg.set_title(_("Change Instrument Type"))
			self.res.get_widget("instructions").set_text(
					  _("Choose the new instrument type for %s") % self.instr.name)
			self.okbutton.set_label("gtk-ok")

		

		self.tree.connect("item-activated", self.OnSelected)
		self.tree.connect("selection-changed", self.OnSelectionChanged)

		self.model = gtk.ListStore(str, str, gtk.gdk.Pixbuf)
			
		for instr in Globals.getCachedInstruments():
			lineList = [x.center(12) for x in textwrap.wrap(instr[0], 11)]
			newList = "\n".join(lineList)
			self.model.append((newList, instr[1], instr[2]))
		
		self.tree.set_model(self.model)
			
		self.tree.set_text_column(0)
		self.tree.set_pixbuf_column(2)
		self.tree.set_orientation(gtk.ORIENTATION_VERTICAL)
		if self.instr:
			self.tree.set_selection_mode(gtk.SELECTION_SINGLE)
		else:
			self.tree.set_selection_mode(gtk.SELECTION_MULTIPLE)
			
		self.tree.set_item_width(90)
		self.tree.set_size_request(72, -1)
				
		self.width = int(Globals.settings.general["addinstrumentwindowwidth"])
		self.height = int(Globals.settings.general["addinstrumentwindowheight"])
		self.dlg.resize(self.width, self.height)

		self.dlg.set_icon(self.parent.icon)
		self.dlg.set_transient_for(self.parent.window)
		self.dlg.show()
	#_____________________________________________________________________
	
	def OnSelected(self, iconview, path):
		"""
		Calls the OnOK method when an instrument has been selected.
		
		Parameters:
			iconview -- reserved for GTK callbacks, don't use it explicitly.
			path -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.OnOK()

	#_____________________________________________________________________
			
	def OnOK(self, button=None):
		"""
		This method is called when the ok button in the dialog has been clicked.
		It will then add the selected instrument into the main jokosher window (JokosherApp).  
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicity.
		"""
		selectedItems = self.tree.get_selected_items()
		if not self.instr:
			instrList = []
			for i in selectedItems:
				item = self.model[i[0]]
				#find the actual instrument using index 1 (the instrument type)
				#because the name has been wrapped in self.model and can't be used
				name, type, pixbuf, path = [x for x in Globals.getCachedInstruments() if x[1] == item[1]][0]
				instrList.append( (name, type) )
			
			self.project.AddInstruments(instrList)
		else:
			item = self.model[selectedItems[0][0]]
			instrItem = [x for x in Globals.getCachedInstruments() if x[1] == item[1]][0]
			self.instr.ChangeType(instrItem[1], instrItem[0])


		self.dlg.destroy()
		
	#_____________________________________________________________________
	
	def OnCancel(self, button):
		"""
		Called when the cancel button in the dialog has been clicked.
		
		Parameters:
			button -- reserved for GTK callbacks, dont't use it explicity.
		"""

		self.dlg.destroy()
		
	#_____________________________________________________________________

	def OnSelectionChanged(self, button):
		"""
		Called when a new instrument icon is chosen.
		The ok button in the dialog will appear inactive if there are no instruments selected.
		
		Parameters:
			button -- reserved for GTK callbacks, dont't use it explicity.
		"""
		sel = self.tree.get_selected_items()

		if len(sel) <= 0:
			self.okbutton.set_sensitive(False)
		else:
			self.okbutton.set_sensitive(True)
			
	#_____________________________________________________________________
	
	def OnSearchChange(self, widget):
		"""
		This method will be called when a new letter has been added to the search box.
		It will then update the search results accordingly.
		
		Parameters:
			widget -- reserved for GTK callbacks, dont't use it explicity.
		"""
		search_text = self.search_entry.get_text()
		self.model = gtk.ListStore(str, str, gtk.gdk.Pixbuf)
		
		for instr in Globals.getCachedInstruments():
			if search_text.lower() in instr[0].lower():
				self.model.append((instr[0], instr[1], instr[2]))
		
		self.tree.set_model(self.model)
		
	#_____________________________________________________________________


	def OnResize(self, widget, event):
		"""
		This method is called when the add instrument dialog is resized

		Parameters:
			widget -- GTK callback parameter.
			event -- GTK callback parameter.
			
		Returns:
			False -- continue GTK signal propagation.
		"""		

		(self.width, self.height) = widget.get_size()

		return False

	#_____________________________________________________________________

	def OnDestroy(self, widget=None, event=None):
		"""
		Called when the add instrument dialog is destroyed

		Parameters: 
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		Globals.settings.general["addinstrumentwindowwidth"] = self.width
		Globals.settings.general["addinstrumentwindowheight"] = self.height
		Globals.settings.write()
		
	#_____________________________________________________________________	



#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	TimeView.py
#	
#	This module holds and updates the gtk.Label which displays
#	the current time position.
#
#-------------------------------------------------------------------------------

import gtk
import gettext
_ = gettext.gettext

#=========================================================================

class TimeView(gtk.EventBox):
	"""
	This class updates the time label which displays the time position of a loaded project.
	"""

	""" GTK widget name """
	__gtype_name__ = 'TimeView'

	#_____________________________________________________________________

	def __init__(self, project):
		"""
		Creates a new instance of TimeView
		
		Parameters:
			project -- reference to Project (Project.py)
		"""
		gtk.EventBox.__init__(self)
		self.set_events(gtk.gdk.BUTTON_PRESS_MASK)
		self.connect("button_press_event", self.OnClick)
		self.timeViewLabel = gtk.Label()
		self.add(self.timeViewLabel)
		self.project = project
		# Listen for bpm and time sig changes
		self.project.connect("bpm", self.OnProjectTime)
		self.project.connect("time-signature", self.OnProjectTime)
		# Listen for playback position and mode changes
		self.project.transport.connect("transport-mode", self.OnTransportMode)
		self.project.transport.connect("position", self.OnTransportPosition)
		
		self.UpdateTime()
		self.tmtip = gtk.Tooltips()
		self.tmtip.set_tip(self, _("Double click to change the time format"), None)
		
	#_____________________________________________________________________
		
	def UpdateTime(self):
		"""
		Updates the time label.
		"""		
		transport = self.project.transport
		formatString = "<span font_desc='Sans Bold 15'>%s</span>"
		if transport.mode == transport.MODE_BARS_BEATS:
			bars, beats, ticks = transport.GetPositionAsBarsAndBeats()
			self.timeViewLabel.set_markup(formatString%("%05d:%d:%03d"%(bars, beats, ticks)))
			
		elif transport.mode == transport.MODE_HOURS_MINS_SECS:
			hours, mins, secs, millis = transport.GetPositionAsHoursMinutesSeconds()
			self.timeViewLabel.set_markup(formatString%("%01d:%02d:%02d:%03d"%(hours, mins, secs, millis)))
			
	#_____________________________________________________________________
	
	def OnProjectTime(self, project):
		"""
		Callback for when the project's time related properties
		(bpm and time signature) change.
		
		Parameters:
			project -- The project instance that send the signal.
		"""
		self.UpdateTime()
	
	#_____________________________________________________________________
	
	def OnTransportMode(self, transportManager, mode):
		"""
		Callback for signal when the transport mode changes.
		
		Parameters:
			transportManager -- the TransportManager instance that send the signal.
			mode -- the mode type that the transport changed to.
		"""
		self.UpdateTime()
	
	#_____________________________________________________________________
	
	def OnTransportPosition(self, transportManager, extraString):
		"""
		Callback for signal when the transport position changes.
		
		Parameters:
			transportManager -- the TransportManager instance that send the signal.
			extraString -- a string specifying the extra action details. i.e. "stop-action"
					means that the position changed because the user hit stop.
		"""
		self.UpdateTime()
	
	#_____________________________________________________________________

	def OnClick(self, widget, event):
		"""
		Called when the label is double clicked.
		It will then change the time label to represent either MODE_HOURS_MINS_SECS or MODE_BARS_BEATS.
		MODE_HOURS_MINS_SECS - time in seconds, minutes and hours.
		MODE_BARS_BEATS - time in how many beats are in each bar.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.	
		"""
		if event.type == gtk.gdk._2BUTTON_PRESS:
			transport = self.project.transport
			if transport.mode == transport.MODE_BARS_BEATS:
				self.project.SetTransportMode(transport.MODE_HOURS_MINS_SECS)
			else:
				self.project.SetTransportMode(transport.MODE_BARS_BEATS)
			
	#_____________________________________________________________________
	
#=========================================================================




import Event
import Utils
import os.path
import xml.dom.minidom as xml

import gettext
_ = gettext.gettext

#=========================================================================

class NewEvent:
	def __init__(self, instr_id, filename, event_start, event_id, recording=False):
		self.instr_id = instr_id
		self.filename = filename
		self.event_start = event_start
		self.id = event_id
		self.recording = recording
		
	def Execute(self, project, event_duration=None, event_levels_file=None):
		instr = project.JokosherObjectFromString("I" + str(self.instr_id))
		
		filename = self.filename
		if not os.path.isabs(filename):
			# If there is a relative path for filename, this means it is in the project's audio dir
			filename = os.path.join(project.audio_path, filename)
		
		if self.recording:
			event_name = _("Recorded audio")
		else:
			event_name = None
			
		instr.addEventFromFile(self.event_start, filename, copyfile=False,
		              name=event_name, duration=event_duration, levels_file=event_levels_file)
		
	def StoreToString(self):
		doc = xml.Document()
		node = doc.createElement("NewEvent")
		doc.appendChild(node)
		
		node.setAttribute("instrument_id", str(self.instr_id))
		node.setAttribute("file", self.filename)
		node.setAttribute("start", str(self.event_start))
		node.setAttribute("event_id", str(self.id))
		node.setAttribute("recording", str(self.recording))
				
		return doc.toxml()
	
	@staticmethod
	def LoadFromString(string):
		doc = xml.parseString(string)
		node = doc.firstChild
		assert node.nodeName == "NewEvent"
		
		instr_id = int(node.getAttribute("instrument_id"))
		filename = node.getAttribute("file")
		event_start = float(node.getAttribute("start"))
		id = int(node.getAttribute("event_id"))
		recording = (node.getAttribute("recording").lower() == "true")
		
		return NewEvent(instr_id, filename, event_start, id, recording)

#=========================================================================

class StartDownload:
	def __init__(self, instr_id, url, save_file, event_start, event_id):
		self.instr_id = instr_id
		self.url = url
		self.save_file = save_file
		self.event_start = event_start
		self.id = event_id
		
	def Execute(self, project):
		instr = project.JokosherObjectFromString("I" + str(self.instr_id))
		instr.addEventFromURL(self.event_start, self.url)
		
	def GetNewEventAction(self):
		return NewEvent(self.instr_id, self.save_file, self.event_start, self.id)
		
	def StoreToString(self):
		doc = xml.Document()
		node = doc.createElement("StartDownload")
		doc.appendChild(node)
		
		node.setAttribute("instrument_id", str(self.instr_id))
		node.setAttribute("url", self.url)
		node.setAttribute("save_file", self.save_file)
		node.setAttribute("start", str(self.event_start))
		node.setAttribute("event_id", str(self.id))
				
		return doc.toxml()
	
	@staticmethod
	def LoadFromString(string):
		doc = xml.parseString(string)
		node = doc.firstChild
		assert node.nodeName == "StartDownload"
		
		instr_id = int(node.getAttribute("instrument_id"))
		url = node.getAttribute("url")
		save_file = node.getAttribute("save_file")
		event_start = float(node.getAttribute("start"))
		id = int(node.getAttribute("event_id"))
		
		return StartDownload(instr_id, url, save_file, event_start, id)

#=========================================================================

class CompleteLoading:
	def __init__(self, event_id, duration, levels_file):
		self.id = event_id
		self.duration = duration
		self.levels_file = levels_file
		
	def Execute(self, project):
		pass
		
	def StoreToString(self):
		doc = xml.Document()
		node = doc.createElement("CompleteLoading")
		doc.appendChild(node)
		node.setAttribute("event_id", str(self.id))
		node.setAttribute("duration", str(self.duration))
		node.setAttribute("levels_file", self.levels_file)
				
		return doc.toxml()
	
	@staticmethod
	def LoadFromString(string):
		doc = xml.parseString(string)
		node = doc.firstChild
		assert node.nodeName == "CompleteLoading"
		
		id = int(node.getAttribute("event_id"))
		duration = float(node.getAttribute("duration"))
		levels_file = node.getAttribute("levels_file")
		
		return CompleteLoading(id, duration, levels_file)

#=========================================================================

class Action:
	def __init__(self, objectString, func_name, args, kwargs):
		self.objectString = objectString
		self.func_name = func_name
		self.args = args
		self.kwargs = kwargs
			
	def Execute(self, project):
		target_object = project.JokosherObjectFromString(self.objectString)
		args = []
		kwargs = {}
		
		for obj in self.args:
			if isinstance(obj, MockEvent):
				obj = project.JokosherObjectFromString(obj.event_string)
			args.append(obj)
			
		for key, value in self.kwargs.iteritems():
			if isinstance(value, MockEvent):
				value = project.JokosherObjectFromString(value.event_string)
			kwargs[key] = value
		
		getattr(target_object, self.func_name)(*args, **kwargs)

		
	def StoreToString(self):
		doc = xml.Document()
		action = doc.createElement("Action")
		doc.appendChild(action)
		
		action.setAttribute("object", self.objectString)
		action.setAttribute("function", self.func_name)
		
		for arg in self.args:
			node = doc.createElement("Argument")
			action.appendChild(node)
			self.WriteToXMLAttributes(None, arg, node)
				
		for key, value in self.kwargs.iteritems():
			node = doc.createElement("NamedArgument")
			action.appendChild(node)
			self.WriteToXMLAttributes(key, value, node)
				
		return doc.toxml()
	
	def WriteToXMLAttributes(self, key, value, node):
		if key:
			node.setAttribute("key", key)
		
		if isinstance(value, Event.Event) or isinstance(value, MockEvent):
			node.setAttribute("type", "Event")
			node.setAttribute("value", str(value.id))
		else:
			Utils.StoreVariableToNode(value, node, "type", "value")
	
	@staticmethod
	def ReadFromXMLAttributes(node):
		type = node.getAttribute("type")
		if type == "Event":
			event_id = node.getAttribute("value")
			value = MockEvent("E" + event_id)
			assert value is not None
		else:
			value = Utils.LoadVariableFromNode(node, "type", "value")
			
		return value
				
	@staticmethod
	def LoadFromString(string):
		doc = xml.parseString(string)
		actionNode = doc.firstChild
		assert actionNode.nodeName == "Action"
		
		function_name = actionNode.getAttribute("function")
		object_string = actionNode.getAttribute("object")
		
		argsList = []
		kwArgsDict = {}
		
		for argNode in actionNode.childNodes:
			value = Action.ReadFromXMLAttributes(argNode)
			if argNode.nodeName == "Argument":
				argsList.append(value)
			elif argNode.nodeName == "NamedArgument":
				key = str(argNode.getAttribute("key"))
				kwArgsDict[key] = value

		return Action(object_string, function_name, argsList, kwArgsDict)

#=========================================================================

class SetNotes:
	def __init__(self, notes):
		self.notes = notes
		
	def Execute(self, project):
		project.SetNotes(self.notes)
		
	def StoreToString(self):
		doc = xml.Document()
		node = doc.createElement("SetNotes")
		doc.appendChild(node)
		
		node.setAttribute("notes", repr(self.notes))
				
		return doc.toxml()
	
	@staticmethod
	def LoadFromString(string):
		doc = xml.parseString(string)
		node = doc.firstChild
		assert node.nodeName == "SetNotes"
		
		notes = Utils.StringUnRepr(node.getAttribute("notes"))
		
		return SetNotes(notes)

#=========================================================================
# Helper functions for Project related Actions

def Undo():
	return Action("P", "Undo", tuple(), dict())

def Redo():
	return Action("P", "Redo", tuple(), dict())

def SetName(name):
	return Action("P", "SetName", (name,), dict())

def SetAuthor(author):
	return Action("P", "SetAuthor", (author,), dict())

def InstrumentSetInput(id, device, inTrack):
	return Action("I" + str(id), "SetInput", (device, inTrack), dict())

def InstrumentSetVolume(id, volume):
	return Action("I" + str(id), "SetVolume", (volume,), dict())

#=========================================================================

def LoadFromString(string):
	doc = xml.parseString(string)
	node = doc.firstChild.nodeName
	
	action_dict = {
		"Action" : Action,
		"NewEvent" : NewEvent,
		"StartDownload" : StartDownload,
		"CompleteLoading" : CompleteLoading,
		"SetNotes" : SetNotes,
	}
	
	if node in action_dict:
		return action_dict[node].LoadFromString(string)
	
	raise AssertionError("Unknown IncrementalSave node " + node)
	
#=========================================================================

def FilterAndExecuteAll(save_action_list, project):
	complete_load_ids = {}
	for action in save_action_list:
		if isinstance(action, CompleteLoading):
			complete_load_ids[action.id] = action
	
	for action in save_action_list:
		loading_type = isinstance(action, StartDownload) or isinstance(action, NewEvent)
		if loading_type and action.id in complete_load_ids:
			# convert to a NewEvent command, skip redownload
			if isinstance(action, StartDownload):
				action = action.GetNewEventAction()
			# loading has completed. Instead of re-downloading or reloading levels
			# just restore from the data on disk
			complete_load = complete_load_ids[action.id]
			action.Execute(project, complete_load.duration, complete_load.levels_file)
		else:
			action.Execute(project)
			
#=========================================================================

class MockEvent:
	def __init__(self, string):
		self.id = int(string[1:])
		self.event_string = string


#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	StatusBar.py
#	
#	This module is a better status bar than the one included
#	with gtk because it allows pango markup (bold, etc.).
#
#-------------------------------------------------------------------------------

import gtk

#=========================================================================

class StatusBar(gtk.Statusbar):
	"""
	Implements an improved status bar which allows pango markup styles (bold, italics, etc).
	"""
	#_____________________________________________________________________

	def __init__(self):
		"""
		Creates a new instance of StatusBar with no messages shown.
		"""
		gtk.Statusbar.__init__(self)
		# gtk.Statusbar contains a label inside a frame inside itself
		self.label = self.get_children()[0].get_children()[0]
		self.label.set_use_markup(True)
		
	#_____________________________________________________________________

	def Push(self, message):
		"""
		Insert a new message into the messages stack.
		
		Parameters:
			message -- string containing the new message to be added to the StatusBar.
			
		Return:
			the value of the next valid message ID.
		"""
		message_id = self.push(0, message)
		self.label.set_use_markup(True)
		return message_id
	
	#_____________________________________________________________________

	def Remove(self, message_id):
		"""
		Removes a new message from the messages stack.
		
		Parameters:
			message_id -- numerical id of the message to be removed from the StatusBar.
		"""
		self.remove(0, message_id)
		self.label.set_use_markup(True)
		
	#_____________________________________________________________________

#=========================================================================





import os, os.path, urllib
import pwd


def getFullName():
	try:
		# Try to get the full name if it exists
		fullname = pwd.getpwuid(os.getuid())[4].split(",")[0]
		if fullname == "":
			fullname = pwd.getpwuid(os.getuid())[0]
		return fullname
	except:
		# If we can't get the fullname, then just use the login
		return pwd.getpwuid(os.getuid())[0]

def samefile(path1, path2):
	return os.path.samefile(path1, path2)


def url2pathname(url):
	return urllib.url2pathname(url)

def pathname2url(path):
	return "file://%s" % urllib.pathname2url(path)

def GetRecordingDefaults():
	return dict()

def GetPlaybackDefaults():
	return dict()




#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Workspace.py
#	
#	A sub-class of gtk.VPaned containing the the varous views of the project
#
#-------------------------------------------------------------------------------

import gtk
import gobject
import os.path
import Globals
import RecordingView
import CompactMixView

#=========================================================================

class Workspace(gtk.VPaned):
	"""
	This class implements the workspace view.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self, project, mainview):
		"""
		Creates a new instance of Workspace.
		
		Parameters:
			project -- the active Project.
			mainview -- reference to the MainApp Jokosher window.
		"""
		gtk.VPaned.__init__(self)
		self.project = project
		self.mainview = mainview
		self.small = False
		self.recordingView = RecordingView.RecordingView(project, mainview, self.small)
		self.mixView = CompactMixView.CompactMixView(project, mainview)
		self.pack1(self.recordingView, resize=True, shrink=False)
		self.pack2(self.mixView, resize=True, shrink=False)
		self.mixView.hide()
		self.show()
	#_____________________________________________________________________
	
	def ToggleCompactMix(self):
		"""
		Toggles compact mix view on/off.
		"""
		if self.mainview.compactMixButton.get_active():
			self.recordingView.ChangeSize(True)
			self.mixView.show()
			self.mainview.compactMixButton.set_tooltip(
						self.mainview.contextTooltips,
						self.mainview.mixingViewEnabledTip)
		else:
			self.recordingView.ChangeSize(False)
			self.mixView.hide()
			self.mainview.compactMixButton.set_tooltip(
						self.mainview.contextTooltips,
						self.mainview.mixingViewDisabledTip)
	#____________________________________________________________________	

#=========================================================================
		



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	This module handles the saving and loading of mixdown profiles.
#	
#
#-------------------------------------------------------------------------------

import os
import gobject
import xml.dom.minidom as xml

import MixdownActions
import Globals
import Utils

#=========================================================================

class MixdownProfiles(gobject.GObject):
	"""
	Represents a mixdown profile, used for doing complex mixdown operations.
	"""
	
	# the extension put on the end of mixdown profiles
	MIXDOWN_EXT = "profile"
	
	"""
	Signals:
		"profile-update" -- MixdownActions have been added to a MixdownProfile.
		"error-occurred" -- An error occurred while attempting to load MixdownActions.
	"""
	
	__gsignals__ = {
		"profile-update" : ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING,) ),
		"error-occurred" : ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING, gobject.TYPE_STRING) )
	}
	
	#_____________________________________________________________________

	def __init__(self, manager):
		""" 
		Creates a new instance of MixdownProfiles.
		
		Parameters:
			manager -- reference to MixdownProfileManager.
		"""
		# we need a reference to MixdownProfileManager because we need to handle
		# special cases for classes such as ExportAsFileType. 
		# ExportAsFileType requires an instance of Project to function correctly,
		# so we have to pass Project and the only way to do that is by accessing the
		# reference to Project in MixdownProfileDialog.
		
		# we also need manager for accessing extensionManager, so we can iterate over
		# loaded extensions, see the __LoadMixdownActionFromAssociatedExtension method for more details.
		self.manager = manager
		
		gobject.GObject.__init__(self)
		
	#_____________________________________________________________________
	
	def SaveProfile(self, name, actionlist=None):
		"""
		This method will write a mixdown profile to JOKOSHER_DATA_HOME/mixdownprofiles/.
		actionlist is a list containing MixdownAction instances, e.g.
		[ExportAsFileType object instance, RunAScript object instance].
	
		Parameters:
			name -- the name of the template file.
			actionlist -- a list containing MixdownAction instances.
		"""
		doc = xml.Document()
		head = doc.createElement("JokosherMixdownProfile")
		doc.appendChild(head)
		actionTag = doc.createElement("MixdownActions")
		head.appendChild(actionTag)
		
		if actionlist:
			for action in actionlist:
				actionName = doc.createElement(action.__class__.__name__)
				actionTag.appendChild(actionName)
				actionConfig = doc.createElement("Configuration")
				actionName.appendChild(actionConfig)
				Utils.StoreDictionaryToXML(doc, actionConfig, action.config)
		
		if not name.endswith("." + self.MIXDOWN_EXT):
			name += "." + self.MIXDOWN_EXT
		namePath = os.path.join(Globals.MIXDOWN_PROFILES_PATH, name)
		try:
			try:
				filename = open(namePath, "w")
				filename.write(doc.toprettyxml())
			except IOError, e:
				Globals.debug(_("The mixdown profile %s does not exist" % namePath))
		finally:
			filename.close()
			
		self.emit("profile-update", "saveProfile")
	
	#_____________________________________________________________________

	def DeleteProfile(self, name):
		"""
		This method will delete a mixdown profile.
		
		Parameters:
			name -- the name of the mixdown profile which will be deleted.
		"""
		if not name.endswith("." + self.MIXDOWN_EXT):
			name += "." + self.MIXDOWN_EXT
		namePath = os.path.join(Globals.MIXDOWN_PROFILES_PATH, name)
		try:
			os.remove(namePath)
		except OSError, e:
			Globals.debug(_("Cannot remove mixdown profile %s" % namePath))
		
		self.emit("profile-update", "deleteProfile")
	
	#_____________________________________________________________________

	def LoadMixdownActionsFromProfile(self, name):
		"""
		Called when actions from a profile need to be loaded.
		This method will return a list of MixdownAction instances associated
		with the profile name specified.
		
		Parameters:
			name-- the name of a profile to load actions from.
		
		Returns:
			actions -- a list of MixdownAction instances assocated with the profile specified.
		"""
		if not name.endswith("." + self.MIXDOWN_EXT):
			name += "." + self.MIXDOWN_EXT
		namePath = os.path.join(Globals.MIXDOWN_PROFILES_PATH, name)
		if os.path.exists(namePath):
			file = open(namePath, "r")
			doc = xml.parse(file)
			actions = []
			for node in doc.getElementsByTagName("MixdownActions"):
				for children in node.childNodes:
					if children.nodeType == 1:
						for child in children.childNodes:
							if child.nodeType == 1:
								actionObject = self.__LoadMixdownAction(children.nodeName)
								if actionObject:
									configActionObject = self.__HandleMixdownActionConfig(actionObject, child)
									actions.append(configActionObject)
			return actions
		else:
			Globals.debug(_("The mixdown profile %s does not exist" % namePath))
		
	#_____________________________________________________________________
	
	def __LoadMixdownAction(self, actionName):
		"""
		Called when a MixdownAction needs to be instantiated.
		This method will return an instance of a MixdownAction specified
		by the action name given.
		
		Parameters:
			actionName -- the name of the MixdownAction to instantiate.
		
		Returns:
			actionObject -- reference to a MixdownAction object.
		"""
		actionObject = None
		
		# load a core MixdownAction, if it fails, actionObject will still be None
		actionObject = self.__LoadMixdownActionFromCore(actionName)
		
		# if we cannot load a core action from the action name specified, then we'll
		# attempt to load it from extensions that have MixdownActions.
		if not actionObject:
			# if __LoadMixdownActionFromAssociatedExtension returns None, then
			# the extension been disabled or deleted.
			actionObject = self.__LoadMixdownActionFromAssociatedExtension(actionName)
		
		# if we have an action object then it is time to instantiate it.
		if actionObject:
			actionObject = self.__HandleSpecialCases(actionObject)
		
		return actionObject

	#_____________________________________________________________________
	
	def __LoadMixdownActionFromCore(self, actionName):
		"""
		Called when a MixdownAction from MixdownActions needs to be instantiated.
		This method returns an instantiated MixdownAction from MixdownActions.py (core)
		
		Parameters:
			actionName -- the name of the MixdownAction to instantiate.
		
		Returns:
			actionObject -- reference to a MixdownAction object.
		"""
		actionObject = None
		try:
			actionObject = getattr(MixdownActions, actionName)
		except AttributeError:
			pass
		
		return actionObject
	
	#_____________________________________________________________________

	def __LoadMixdownActionFromAssociatedExtension(self, actionName):
		"""
		Called when a MixdownAction from a Jokosher Extension needs to be instantiated.
		This method will iterate over all loaded extensions and return the MixdownAction
		object specified by the action name given.
		
		Parameters:
			actionName -- the name of the MixdownAction to instantiate.
		
		Returns:
			action -- reference to a MixdownAction object.
		"""
		count = -1
		action = None
		for extension in self.__ReturnLoadedExtensions():
			try:
				if extension["enabled"] and extension["extension"].mixdownActions:
					for item in extension["extension"].mixdownActions:
						count += 1
						if actionName == item.__name__:
							action = extension["extension"].mixdownActions[count]
							break
				else:
					self.__RaiseMixdownActionLoadingError(actionName, extension["name"])
			except AttributeError:
				pass

		return action
	
	#_____________________________________________________________________
	
	def __ReturnLoadedExtensions(self):
		"""
		Called when all loaded extension need to be returned.
		This method returns all loaded Jokosher extensions.

		Returns:
			loadedExtensions -- list of loaded Jokosher extensions.
		"""
		return self.manager.mixdownProfileDialog.mainapp.extensionManager.loadedExtensions

	#_____________________________________________________________________
	
	def __HandleMixdownActionConfig(self, actionObject, configTag):
		"""
		Called when the configuration of a MixdownAction needs to be set.
		This method handles setting the isConfigured class attribute for
		MixdownActions.
		
		Parameters:
			actionObject -- reference to a MixdownAction object.
			actionTag -- the action tag to load the config from, e.g. <ExportAsFileType> <Configuration/>
		
		Returns:
			actionObject -- the configured MixdownAction
		"""
		actionObject.config = Utils.LoadDictionaryFromXML(configTag)
		
		if actionObject.config:
			actionObject.isConfigured = True
		else:
			actionObject.isConfigured = False
		
		return actionObject
		
	#_____________________________________________________________________

	def __RaiseMixdownActionLoadingError(self, actionName, extensionName):
		"""
		Called when an error has occured while loading MixdownActions.
		This method will emit an error-occurred gobject signal, this signal
		will show a dialog, informing the user that a MixdownAction could not
		be loaded.
		
		Parameters:
			actionName -- the name of the MixdownAction which cannot be loaded.
			extensionName -- the name of the extension that the MixdownAction can't be loaded from.
		"""
		self.emit("error-occurred", actionName, extensionName)

	#_____________________________________________________________________
	
	def __HandleSpecialCases(self, actionObject):
		"""
		Called when special cases such as ExportAsFileType need to be handled.
		This method will take care of the loading of MixdownActions from MixdownActions.py,
		passing the correct arguments to the MixdownAction classes,
		e.g. ExportAsFileType requires Project to be passed to it's constructor.
		
		Parameters:
			actionObject -- reference to a MixdownAction object.
		
		Returns:
			actionObject - a MixdownAction instance.
		"""
		if actionObject == MixdownActions.ExportAsFileType:
			action = actionObject(self.manager.mixdownProfileDialog.project)
		else:
			action = actionObject()
		return action
		
	#_____________________________________________________________________

#=========================================================================



# This file is part of the Jokosher project.
# Original copied from PiTiVi , Non-linear video editor.
#
#       elements/singledecodebin.py
#
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
# Re-licensing to GPL 2 or later granted by Edward Hervey
# on February 14th, 2008.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


"""
Single-stream queue-less decodebin
"""

import gobject
import gst

def is_raw(caps):
    """ returns True if the caps are RAW """
    rep = caps.to_string()
    valid = ["video/x-raw", "audio/x-raw", "text/plain", "text/x-pango-markup"]
    for val in valid:
        if rep.startswith(val):
            return True
    return False

class SingleDecodeBin(gst.Bin):
    """
    A variant of decodebin.

    * Only outputs one stream
    * Doesn't contain any internal queue
    """

    __gsttemplates__ = (
        gst.PadTemplate ("sinkpadtemplate",
                         gst.PAD_SINK,
                         gst.PAD_ALWAYS,
                         gst.caps_new_any()),
        gst.PadTemplate ("srcpadtemplate",
                         gst.PAD_SRC,
                         gst.PAD_SOMETIMES,
                         gst.caps_new_any())
        )
    def __init__(self, caps=None, uri=None, *args, **kwargs):
        gst.Bin.__init__(self, *args, **kwargs)
        if not caps:
            caps = gst.caps_new_any()
        self.caps = caps
        self.typefind = gst.element_factory_make("typefind", "internal-typefind")
        self.add(self.typefind)

        self.uri = uri
        if self.uri and gst.uri_is_valid(self.uri):
            self.urisrc = gst.element_make_from_uri(gst.URI_SRC, uri, "urisrc")
            self.log("created urisrc %s / %r" % (self.urisrc.get_name(),
                                                 self.urisrc))
            self.add(self.urisrc)
            self.urisrc.link(self.typefind)
        else:
            self._sinkpad = gst.GhostPad("sink", self.typefind.get_pad("sink"))
            self._sinkpad.set_active(True)
            self.add_pad(self._sinkpad)

        self.typefind.connect("have_type", self._typefindHaveTypeCb)

        self._srcpad = None

        self._dynamics = []

        self._validelements = [] #added elements

        self._factories = self._getSortedFactoryList()


    ## internal methods

    def _controlDynamicElement(self, element):
        self.log("element:%s" % element.get_name())
        self._dynamics.append(element)
        element.connect("pad-added", self._dynamicPadAddedCb)
        element.connect("no-more-pads", self._dynamicNoMorePadsCb)

    def _getSortedFactoryList(self):
        """
        Returns the list of demuxers, decoders and parsers available, sorted
        by rank
        """
        def myfilter(fact):
            if fact.get_rank() < 64 :
                return False
            klass = fact.get_klass()
            if not ("Demuxer" in klass or "Decoder" in klass or "Parse" in klass):
                return False
            return True
        reg = gst.registry_get_default()
        res = [x for x in reg.get_feature_list(gst.ElementFactory) if myfilter(x)]
        res.sort(lambda a, b: int(b.get_rank() - a.get_rank()))
        return res

    def _findCompatibleFactory(self, caps):
        """
        Returns a list of factories (sorted by rank) which can take caps as
        input. Returns empty list if none are compatible
        """
        self.debug("caps:%s" % caps.to_string())
        res = []
        for factory in self._factories:
            for template in factory.get_static_pad_templates():
                if template.direction == gst.PAD_SINK:
                    intersect = caps.intersect(template.static_caps.get())
                    if not intersect.is_empty():
                        res.append(factory)
                        break
        self.debug("returning %r" % res)
        return res

    def _closeLink(self, element):
        """
        Inspects element and tries to connect something on the srcpads.
        If there are dynamic pads, it sets up a signal handler to
        continue autoplugging when they become available.
        """
        to_connect = []
        dynamic = False
        templates = element.get_pad_template_list()
        for template in templates:
            if not template.direction == gst.PAD_SRC:
                continue
            if template.presence == gst.PAD_ALWAYS:
                pad = element.get_pad(template.name_template)
                to_connect.append(pad)
            elif template.presence == gst.PAD_SOMETIMES:
                pad = element.get_pad(template.name_template)
                if pad:
                    to_connect.append(pad)
                else:
                    dynamic = True
            else:
                self.log("Template %s is a request pad, ignoring" % pad.name_template)

        if dynamic:
            self.debug("%s is a dynamic element" % element.get_name())
            self._controlDynamicElement(element)

        for pad in to_connect:
            self._closePadLink(element, pad, pad.get_caps())

    def _tryToLink1(self, source, pad, factories):
        """
        Tries to link one of the factories' element to the given pad.

        Returns the element that was successfully linked to the pad.
        """
        self.debug("source:%s, pad:%s , factories:%r" % (source.get_name(),
                                                         pad.get_name(),
                                                         factories))
        result = None
        for factory in factories:
            element = factory.create()
            if not element:
                self.warning("weren't able to create element from %r" % factory)
                continue

            sinkpad = element.get_pad("sink")
            if not sinkpad:
                continue

            self.add(element)
            try:
                pad.link(sinkpad)
            except:
                element.set_state(gst.STATE_NULL)
                self.remove(element)
                continue

            self._closeLink(element)
            element.set_state(gst.STATE_PAUSED)

            result = element
            break

        return result

    def _closePadLink(self, element, pad, caps):
        """
        Finds the list of elements that could connect to the pad.
        If the pad has the desired caps, it will create a ghostpad.
        If no compatible elements could be found, the search will stop.
        """
        self.debug("element:%s, pad:%s, caps:%s" % (element.get_name(),
                                                    pad.get_name(),
                                                    caps.to_string()))
        if caps.is_empty():
            self.log("unknown type")
            return
        if caps.is_any():
            self.log("type is not know yet, waiting")
            return
        if caps.intersect(self.caps):
            # This is the desired caps
            if not self._srcpad:
                self._wrapUp(element, pad)
        elif is_raw(caps):
            self.log("We hit a raw caps which isn't the wanted one")
            # FIXME : recursively remove everything until demux/typefind

        else:
            # Find something
            if len(caps) > 1:
                self.log("many possible types, delaying")
                return
            facts = self._findCompatibleFactory(caps)
            if not facts:
                self.log("unknown type")
                return
            self._tryToLink1(element, pad, facts)

    def _wrapUp(self, element, pad):
        """
        Ghost the given pad of element.
        Remove non-used elements.
        """

        if self._srcpad:
            return
        self._markValidElements(element)
        self._removeUnusedElements(self.typefind)
        self.log("ghosting pad %s" % pad.get_name())
        self._srcpad = gst.GhostPad("src", pad)
        self._srcpad.set_active(True)
        self.add_pad(self._srcpad)
        self.post_message(gst.message_new_state_dirty(self))

    def _markValidElements(self, element):
        """
        Mark this element and upstreams as valid
        """
        self.log("element:%s" % element.get_name())
        if element == self.typefind:
            return
        self._validelements.append(element)
        # find upstream element
        pad = list(element.sink_pads())[0]
        parent = pad.get_peer().get_parent()
        self._markValidElements(parent)

    def _removeUnusedElements(self, element):
        """
        Remove unused elements connected to srcpad(s) of element
        """
        self.log("element:%r" % element)
        for pad in element.src_pads():
            if pad.is_linked():
                peer = pad.get_peer().get_parent()
                self._removeUnusedElements(peer)
                if not peer in self._validelements:
                    self.log("removing %s" % peer.get_name())
                    pad.unlink(pad.get_peer())
                    peer.set_state(gst.STATE_NULL)
                    self.remove(peer)

    def _cleanUp(self):
        self.log("")
        if self._srcpad:
            self.remove_pad(self._srcpad)
        self._srcpad = None
        for element in self._validelements:
            element.set_state(gst.STATE_NULL)
            self.remove(element)
        self._validelements = []

    ## Overrides

    def do_change_state(self, transition):
        self.debug("transition:%r" % transition)
        res = gst.Bin.do_change_state(self, transition)
        if transition in [gst.STATE_CHANGE_PAUSED_TO_READY, gst.STATE_CHANGE_READY_TO_NULL]:
            self._cleanUp()
        return res

    ## Signal callbacks

    def _typefindHaveTypeCb(self, typefind, probability, caps):
        self.debug("probability:%d, caps:%s" % (probability, caps.to_string()))
        self._closePadLink(typefind, typefind.get_pad("src"), caps)

    ## Dynamic element Callbacks

    def _dynamicPadAddedCb(self, element, pad):
        self.log("element:%s, pad:%s" % (element.get_name(), pad.get_name()))
        if not self._srcpad:
            self._closePadLink(element, pad, pad.get_caps())

    def _dynamicNoMorePadsCb(self, element):
        self.log("element:%s" % element.get_name())

gobject.type_register(SingleDecodeBin)



#!/usr/bin/python

#
#    THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#    THE 'COPYING' FILE FOR DETAILS
#
#    This module is meant for testing and profiling the code only.
#    This file should not be included in any release.
#
#-------------------------------------------------------------------------------

import hotshot
from hotshot import stats

import JokosherApp

profile = hotshot.Profile("JokosherApp", lineevents=1)
profile.runcall(JokosherApp.main)

s = stats.load("JokosherApp")

s.strip_dirs()
s.sort_stats("cumulative", "calls").print_stats()
s.sort_stats("time", "calls").print_stats()




#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	This class handles all of the processing associated with the
#	Instrument Connections dialog.
#
#-------------------------------------------------------------------------------

import gtk.glade
import gobject
import pygst
pygst.require("0.10")
import gst
import Globals
import AudioBackend
import PreferencesDialog
import gettext
_ = gettext.gettext

#=========================================================================

class InstrumentConnectionsDialog:
	"""
	Handles all of the processing associated with the Instrument Connections dialog.
	"""
	
	#_____________________________________________________________________

	def __init__(self, project, parent):
		"""
		Creates a new instance of InstrumentConnectionsDialog.
		
		Parameters:
			project -- the currently active Project.
			parent -- reference to the MainApp Jokosher window.
		"""
		if project:
			self.project = project
		else:
			return
		
		self.res = gtk.glade.XML(Globals.GLADE_PATH, "InstrumentConnectionsDialog")

		self.signals = {
			"on_close_clicked" : self.OnClose,
			"on_change_sound_system" : self.OnChangeSoundSystem,
		}
		
		self.res.signal_autoconnect(self.signals)

		self.window = self.res.get_widget("InstrumentConnectionsDialog")
		self.vbox = self.res.get_widget("vbox")
		
		if len(self.project.instruments) > 0:
			self.Populate()
		else:
			self.res.get_widget("explainLabel").set_text(_("There are no instruments to connect"))

		self.parent = parent
		self.window.set_icon(self.parent.icon)
		## centre the InstrumentConnectionsDialog on the main jokosher window
		self.window.set_transient_for(self.parent.window)
		self.window.show_all()

	#_____________________________________________________________________

	def OnClose(self, button):
		"""
		Called when the dialog gets closed.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.window.destroy()
	
	#_____________________________________________________________________
	
	def OnChangeSoundSystem(self, button):
		prefsdlg = PreferencesDialog.PreferencesDialog(self.project, self.parent, self.parent.icon)
		self.window.hide()
		# TODO: don't create a new instance of this window each time
		prefsdlg.dlg.connect("destroy", self.parent.OnInstrumentConnectionsDialog)
		prefsdlg.dlg.show_all()
	
	#_____________________________________________________________________

	def Populate(self):
		"""
		Creates all the widgets for intruments and devices that compose the
		InstrumentConnectionsDialog and then adds them to the dialog.
		"""
		
		self.devices_list = []
		liststore = gtk.ListStore(gobject.TYPE_STRING)
	
		#Find out how many channels a device offers
		for device, deviceName in AudioBackend.ListCaptureDevices():
			#Don't want the default device twice (once as 'default' and once as its actual hw ref)
			# Default will always be the first one, and have no name.
			if not self.devices_list and not deviceName:
				if device == "default":
					display = _("Default")
				else:
					display = _("Default (%s)") % device
				self.devices_list.append((device, display, -1))
				liststore.append((display,))
			else:			
				num_channels = AudioBackend.GetChannelsOffered(device)
				for input in xrange(num_channels):
					if num_channels > 1:
						s = _("%(device)s (%(id)s), input %(input)d")
						display = s % {"device":deviceName, "id":device, "input":input}
					else:
						display = _("%(device)s (%(id)s)") % {"device":deviceName, "id":device}
					self.devices_list.append((device, deviceName, input))
					liststore.append((display,))
		
		if self.devices_list:
			for instr in self.project.instruments:
				instrument = instr
				row = gtk.HBox()
				row.set_spacing(10)
				image = gtk.Image()
				image.set_from_pixbuf(instrument.pixbuf)
				label = gtk.Label(instrument.name)
				
				combobox = gtk.ComboBox(liststore)
				cell = gtk.CellRendererText()
				combobox.pack_start(cell, True)
				combobox.add_attribute(cell, 'text', 0)
				
				if instr.input is None:
					# None means default; default is first in combobox
					combobox.set_active(0)
				else:
					currentItem = 0
					for device, deviceName, input in self.devices_list:
						if instr.input == device and input == instr.inTrack:
							combobox.set_active(currentItem)
						currentItem += 1
				
				combobox.connect("changed", self.OnSelected, instr)
				row.pack_start(combobox, True, True)
				row.pack_start(image, False, False)
				row.pack_start(label, False, False)
				
				self.vbox.add(row)
		else:
			audiosrc = Globals.settings.recording["audiosrc"]
			sound_system = None
			for name, element in Globals.CAPTURE_BACKENDS:
				if element == audiosrc:
					sound_system = name
			
			if sound_system:
				msg = _("The %(sound-system-name)s sound system does not support device selection.")
				msg %= {"sound-system-name" : sound_system}
			else:
				msg = _('The "%(custom-pipeline)s" sound system does not support device selection.')
				msg %= {"custom-pipeline" : audiosrc}
			self.res.get_widget("explainLabel").set_text(msg)
			self.vbox.hide()
	
	#_____________________________________________________________________
			
	def OnSelected(self, widget, instr):
		"""
		Sets the instrument's input device.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			instr -- Instrument to change the input device to.
		"""
		device, deviceName, inTrack = self.devices_list[widget.get_active()]
		instr.SetInput(device, inTrack)
			
	#_____________________________________________________________________
	
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	ProjectManager.py
#
#	Contains various helper classes for the Project class as well as
#	all of the loading and saving to and from project files.
#
#=========================================================================

import urlparse, os, gzip, shutil, gst
import Globals, Utils, UndoSystem, LevelsList, IncrementalSave
import Project, Instrument, Event
import xml.dom.minidom as xml
import traceback
import PlatformUtils

def CreateNewProject(projecturi, name, author):
	"""
	Creates a new Project.

	Parameters:
		projecturi -- the filesystem location for the new Project.
						Currently, only file:// URIs are considered valid.
		name --	the name of the Project.
		author - the name of the Project's author.
		
	Returns:
		the newly created Project object.
	"""
	
	if name == "" or author == "" or projecturi == "":
		raise CreateProjectError(4)

	project = InitProjectLocation(projecturi, name)
	project.name = name
	project.author = author
	
	project.SaveProjectFile(project.projectfile)
	return project
	
#_____________________________________________________________________
	
def InitProjectLocation(projecturi, name, project=None):
	"""
	Initialises the folder structure on disk for a Project.
	If no project is provided, a new one is created.
	Otherwise the given project is essentially moved to the new location.

	Parameters:
		projecturi -- the filesystem location for the new Project.
						Currently, only file:// URIs are considered valid.
		name --	the name of the Project.
		project -- the project to init, or None is a new project should be created.
		
	Returns:
		the given Project, or the newly created Project object.
	"""
	if name == "" or projecturi == "":
		raise CreateProjectError(4)
	
	(scheme, domain,folder, params, query, fragment) = urlparse.urlparse(projecturi, "file", False)

	folder = PlatformUtils.url2pathname(folder)

	if scheme != "file":
		# raise "The URI scheme used is invalid." message
		raise CreateProjectError(5)

	filename = name + ".jokosher"
	projectdir = os.path.join(folder, name)

	if not project:
		try:
			project = Project.Project()
		except gst.PluginNotFoundError, e:
			Globals.debug("Missing Gstreamer plugin:", e)
			raise CreateProjectError(6, str(e))
		except Exception, e:
			Globals.debug("Could not initialize project object:", e)
			raise CreateProjectError(1)

	project.projectfile = os.path.join(projectdir, filename)
	project.audio_path = os.path.join(projectdir, "audio")
	project.levels_path = os.path.join(projectdir, "levels")

	if os.path.exists(projectdir):
		raise CreateProjectError(2)
	else: 
		try:
			os.mkdir(projectdir)
			os.mkdir(project.audio_path)
			os.mkdir(project.levels_path)
		except:
			raise CreateProjectError(3)

	return project

#_____________________________________________________________________

def ValidateProject(project):
	"""
	Checks that the Project is valid - i.e. that the files and 
	images it references can be found.
	
	Parameters:
		project -- The project to validate.
	
	Returns:
		True -- the Project is valid.
		False -- the Project contains non-existant files and/or images.
	"""
	unknownfiles=[]
	unknownimages=[]

	for instr in project.instruments:
		for event in instr.events:
			file = event.GetAbsFile()
			if file and (not os.path.exists(file)) and (not file in unknownfiles):
				unknownfiles.append(file)
	if len(unknownfiles) > 0 or len(unknownimages) > 0:
		raise InvalidProjectError(unknownfiles,unknownimages)

	return True
	
#_____________________________________________________________________

def LoadProjectFile(uri):
	"""
	Loads a Project from a saved file on disk.

	Parameters:
		uri -- the filesystem location of the Project file to load. 
				Currently only file:// URIs are considered valid.
				
	Returns:
		the loaded Project object.
	"""
	
	(scheme, domain, projectfile, params, query, fragment) = urlparse.urlparse(uri, "file", False)
	if scheme != "file":
		# raise "The URI scheme used is invalid." message
		raise OpenProjectError(1, scheme)

	projectfile = PlatformUtils.url2pathname(projectfile)

	Globals.debug("Attempting to open:", projectfile)

	if not os.path.exists(projectfile):
		raise OpenProjectError(4, projectfile)

	try:
		try:
			gzipfile = gzip.GzipFile(projectfile, "r")
			doc = xml.parse(gzipfile)
		except IOError, e:
			if e.message == "Not a gzipped file":
				# starting from 0.10, we accept both gzipped xml and plain xml
				file_ = open(projectfile, "r")
				doc = xml.parse(file_)
			else:
				raise e
	except Exception, e:
		Globals.debug(e.__class__, e)
		# raise "This file doesn't unzip" message
		raise OpenProjectError(2, projectfile)
	
	project = Project.Project()
	project.projectfile = projectfile
	projectdir = os.path.split(projectfile)[0]
	project.audio_path = os.path.join(projectdir, "audio")
	project.levels_path = os.path.join(projectdir, "levels")
	try:
		if not os.path.exists(project.audio_path):
			os.mkdir(project.audio_path)
		if not os.path.exists(project.levels_path):
			os.mkdir(project.levels_path)
	except OSError:
		raise OpenProjectError(0)
	
	#only open projects with the proper version number
	version = None
	if doc and doc.firstChild:
		version = doc.firstChild.getAttribute("version")

	if JOKOSHER_VERSION_FUNCTIONS.has_key(version):
		loaderClass = JOKOSHER_VERSION_FUNCTIONS[version]
		Globals.debug("Loading project file version", version)
		try:
			loaderClass(project, doc)
		except:
			tb = traceback.format_exc()
			Globals.debug("Loading project failed", tb)
			raise OpenProjectError(5, tb)
			
		if version != Globals.VERSION:
			#if we're loading an old version copy the project so that it is not overwritten when the user clicks save
			withoutExt = os.path.splitext(projectfile)[0]
			shutil.copy(projectfile, "%s.%s.jokosher" % (withoutExt, version))
		
		project.projectfile = projectfile
		return project
	else:
		# raise a "this project was created in an incompatible version of Jokosher" message
		raise OpenProjectError(3, version)

#=========================================================================

class _LoadZPOFile:
	def __init__(self, project, xmlDoc):
		"""
		Loads a project from a Jokosher 0.1 (Zero Point One) Project file into
		the given Project object using the given XML document.
		
		Parameters:
			project -- the Project instance to apply loaded properties to.
			xmlDoc -- the XML file document to read data from.
		"""
		params = xmlDoc.getElementsByTagName("Parameters")[0]
		
		Utils.LoadParametersFromXML(project, params)
		
		for instr in xmlDoc.getElementsByTagName("Instrument"):
			try:
				id = int(instr.getAttribute("id"))
			except ValueError:
				id = None
			i = Instrument.Instrument(project, None, None, None, id)
			self.LoadInstrument(i, instr)
			project.instruments.append(i)
			if i.isSolo:
				project.soloInstrCount += 1
	
	#_____________________________________________________________________
	
	def LoadInstrument(self, instr, xmlNode):
		"""
		Loads instrument properties from a Jokosher 0.1 XML node
		and saves them to the given Instrument instance.
		
		Parameters:
			instr -- the Instrument instance to apply loaded properties to.
			xmlNode -- the XML node to retreive data from.
		"""
		params = xmlNode.getElementsByTagName("Parameters")[0]
		Utils.LoadParametersFromXML(instr, params)
		#work around because in >0.2 instr.effects is a list not a string.
		instr.effects = []
		
		for ev in xmlNode.getElementsByTagName("Event"):
			try:
				id = int(ev.getAttribute("id"))
			except ValueError:
				id = None
			e = Event.Event(instr, None, id)
			self.LoadEvent(e, ev)
			e.levels_file = e.GetFilename() + Event.Event.LEVELS_FILE_EXTENSION
			instr.events.append(e)
		
		pixbufFilename = os.path.basename(instr.pixbufPath)
		instr.instrType = os.path.splitext(pixbufFilename)[0]
			
		instr.pixbuf = Globals.getCachedInstrumentPixbuf(instr.instrType)
		if not instr.pixbuf:
			Globals.debug("Error, could not load image:", instr.instrType)
			
		#initialize the actuallyIsMuted variable
		instr.OnMute()
		
	#_____________________________________________________________________
	
	def LoadEvent(self, event, xmlNode):
		"""
		Loads event properties from a Jokosher 0.1 XML node
		and saves then to the given Event instance.
		
		Parameters:
			event -- the Event instance to apply loaded properties to.
			xmlNode -- the XML node to retreive data from.
		"""
		params = xmlNode.getElementsByTagName("Parameters")[0]
		Utils.LoadParametersFromXML(event, params)
		
		try:
			xmlPoints = xmlNode.getElementsByTagName("FadePoints")[0]
		except IndexError:
			Globals.debug("Missing FadePoints in Event XML")
		else:
			for n in xmlPoints.childNodes:
				if n.nodeType == xml.Node.ELEMENT_NODE:
					pos = float(n.getAttribute("position"))
					value = float(n.getAttribute("fade"))
					event._Event__fadePointsDict[pos] = value
		
		event.GenerateWaveform()
		event._Event__UpdateAudioFadePoints()
		event.CreateFilesource()
	
	#_____________________________________________________________________

#=========================================================================

class _LoadZPTFile:
	def __init__(self, project, xmlDoc):
		"""
		Loads a Jokosher version 0.2 (Zero Point Two) Project file into
		the given Project object using the given XML document.
		
		Parameters:
			project -- the Project instance to apply loaded properties to.
			xmlDoc -- the XML file document to read data from.
		"""
		self.project = project
		self.xmlDoc = xmlDoc
		
		params = self.xmlDoc.getElementsByTagName("Parameters")[0]
		
		Utils.LoadParametersFromXML(self.project, params)
		
		# Hack to set the transport mode
		self.project.transport.SetMode(self.project.transportMode)
		
		undoRedo = (("Undo", self.project._Project__savedUndoStack),  
		("Redo", self.project._Project__redoStack))  
		for tagName, stack in undoRedo:  
			try:  
				undo = self.xmlDoc.getElementsByTagName(tagName)[0]  
			except IndexError:  
				Globals.debug("No saved %s in project file" % tagName)  
			else:  
				for cmdNode in undo.childNodes:  
					if cmdNode.nodeName == "Command":  
						objectString = str(cmdNode.getAttribute("object"))  
						functionString = str(cmdNode.getAttribute("function"))  
						paramList = Utils.LoadListFromXML(cmdNode)  
					
						functionString = ApplyUndoCompat(objectString, functionString, "0.2")
						
						undoAction = UndoSystem.AtomicUndoAction()  
						undoAction.AddUndoCommand(objectString, functionString, paramList)  
						stack.append(undoAction)  
		
		for instrElement in self.xmlDoc.getElementsByTagName("Instrument"):
			try:
				id = int(instrElement.getAttribute("id"))
			except ValueError:
				id = None
			instr = Instrument.Instrument(self.project, None, None, None, id)
			self.LoadInstrument(instr, instrElement)
			self.project.instruments.append(instr)
			if instr.isSolo:
				self.project.soloInstrCount += 1
				
		for instrElement in self.xmlDoc.getElementsByTagName("DeadInstrument"):
			try:
				id = int(instrElement.getAttribute("id"))
			except ValueError:
				id = None
			instr = Instrument.Instrument(self.project, None, None, None, id)
			self.LoadInstrument(instr, instrElement)
			self.project.graveyard.append(instr)
			instr.RemoveAndUnlinkPlaybackbin()
	
	#_____________________________________________________________________
	
	def LoadInstrument(self, instr, xmlNode):
		"""
		Restores an Instrument from version 0.2 XML representation.
		
		Parameters:
			instr -- the Instrument instance to apply loaded properties to.
			xmlNode -- the XML node to retreive data from.
		"""
		params = xmlNode.getElementsByTagName("Parameters")[0]
		
		Utils.LoadParametersFromXML(instr, params)
		
		globaleffect = xmlNode.getElementsByTagName("GlobalEffect")
		
		for effect in globaleffect:
			elementname = str(effect.getAttribute("element"))
			Globals.debug("Loading effect:", elementname)
			gstElement = instr.AddEffect(elementname)
			
			propsdict = Utils.LoadDictionaryFromXML(effect)
			for key, value in propsdict.iteritems():
				gstElement.set_property(key, value)		
			
		for ev in xmlNode.getElementsByTagName("Event"):
			try:
				id = int(ev.getAttribute("id"))
			except ValueError:
				id = None
			event = Event.Event(instr, None, id)
			self.LoadEvent(event, ev)
			event.levels_file = event.GetFilename() + Event.Event.LEVELS_FILE_EXTENSION
			instr.events.append(event)
		
		for ev in xmlNode.getElementsByTagName("DeadEvent"):
			try:
				id = int(ev.getAttribute("id"))
			except ValueError:
				id = None
			event = Event.Event(instr, None, id)
			self.LoadEvent(event, ev, True)
			event.levels_file = event.GetFilename() + Event.Event.LEVELS_FILE_EXTENSION
			instr.graveyard.append(event)


		#load image from file based on unique type
		instr.pixbuf = Globals.getCachedInstrumentPixbuf(instr.instrType)
		if not instr.pixbuf:
			Globals.debug("Error, could not load image:", instr.instrType)
		
		# load pan level
		instr.panElement.set_property("panorama", instr.pan)
		#check if instrument is muted and setup accordingly
		instr.OnMute()
		#update the volume element with the newly loaded value
		instr.UpdateVolume()
		
	#_____________________________________________________________________
		
	def LoadEvent(self, event, xmlNode, isDead=False):
		"""
		Restores an Event from its version 0.2 XML representation.
		
		Parameters:
			event -- the Event instance to apply loaded properties to.
			xmlNode -- the XML node to retreive data from.
		"""
		params = xmlNode.getElementsByTagName("Parameters")[0]
		
		Utils.LoadParametersFromXML(event, params)
		
		try:
			xmlPoints = xmlNode.getElementsByTagName("FadePoints")[0]
		except IndexError:
			Globals.debug("Missing FadePoints in Event XML")
		else:
			event._Event__fadePointsDict = Utils.LoadDictionaryFromXML(xmlPoints)

		if not isDead:
			#if event.isLoading or event.isRecording:
			# we have to always generate waveform because 0.10 uses different levels format
			event.GenerateWaveform()
			event._Event__UpdateAudioFadePoints()
			event.CreateFilesource()
	
	#_____________________________________________________________________
#=========================================================================

class _LoadZPNFile(_LoadZPTFile):
	LOADING_VERSION = "0.9"
	
	def __init__(self, project, xmlDoc):
		"""
		Loads a Jokosher version 0.9 (Zero Point Nine) Project file into
		the given Project object using the given XML document.
		
		Parameters:
			project -- the Project instance to apply loaded properties to.
			xmlDoc -- the XML file document to read data from.
		"""
		self.project = project
		self.xmlDoc = xmlDoc
		
		params = self.xmlDoc.getElementsByTagName("Parameters")[0]
		
		Utils.LoadParametersFromXML(self.project, params)
		
		notesNode = self.xmlDoc.getElementsByTagName("Notes")
		if notesNode:
			notes = notesNode[0].getAttribute("text")
			# notes are encoded using repr() to preserver \n and \t.
			self.project.notes = Utils.StringUnRepr(notes)
		
		# Hack to set the transport mode
		self.project.transport.SetMode(self.project.transportMode)
		
		undoRedo = (("Undo", self.project._Project__savedUndoStack),
				("Redo", self.project._Project__redoStack))
		for tagName, stack in undoRedo:
			try:
				undo = self.xmlDoc.getElementsByTagName(tagName)[0]
			except IndexError:
				Globals.debug("No saved %s in project file" % tagName)
			else:
				for actionNode in undo.childNodes:
					if actionNode.nodeName == "Action":
						action = UndoSystem.AtomicUndoAction()
						self.LoadUndoAction(action, actionNode)
						stack.append(action)
		
		for instrElement in self.xmlDoc.getElementsByTagName("Instrument"):
			try:
				id = int(instrElement.getAttribute("id"))
			except ValueError:
				id = None
			instr = Instrument.Instrument(self.project, None, None, None, id)
			self.LoadInstrument(instr, instrElement)
			self.project.instruments.append(instr)
			if instr.isSolo:
				self.project.soloInstrCount += 1
		
		for instrElement in self.xmlDoc.getElementsByTagName("DeadInstrument"):
			try:
				id = int(instrElement.getAttribute("id"))
			except ValueError:
				id = None
			instr = Instrument.Instrument(self.project, None, None, None, id)
			self.LoadInstrument(instr, instrElement)
			self.project.graveyard.append(instr)
			instr.RemoveAndUnlinkPlaybackbin()
	
	#_____________________________________________________________________

	def LoadUndoAction(self, undoAction, xmlNode):
		"""
		Loads an AtomicUndoAction from an XML node.
		
		Parameters:
			undoAction -- the AtomicUndoAction instance to save the loaded commands to.
			node -- XML node from which the AtomicUndoAction is loaded.
					Should be an "<Action>" node.
			
		Returns:
			the loaded AtomicUndoAction object.
		"""
		for cmdNode in xmlNode.childNodes:
			if cmdNode.nodeName == "Command":
				objectString = str(cmdNode.getAttribute("object"))
				functionString = str(cmdNode.getAttribute("function"))
				paramList = Utils.LoadListFromXML(cmdNode)
				
				functionString = ApplyUndoCompat(objectString, functionString, self.LOADING_VERSION)
				
				undoAction.AddUndoCommand(objectString, functionString, paramList)
		
	#_____________________________________________________________________
#=========================================================================

class _LoadZPTenFile(_LoadZPNFile):
	LOADING_VERSION = "0.10"
	
	def LoadEvent(self, event, xmlNode, isDead=False):
		"""
		Restores an Event from its version 0.10 XML representation.
		
		Parameters:
			event -- the Event instance to apply loaded properties to.
			xmlNode -- the XML node to retreive data from.
		"""
		params = xmlNode.getElementsByTagName("Parameters")[0]
		
		Utils.LoadParametersFromXML(event, params)
		
		try:
			xmlPoints = xmlNode.getElementsByTagName("FadePoints")[0]
		except IndexError:
			Globals.debug("Missing FadePoints in Event XML")
		else:
			event._Event__fadePointsDict = Utils.LoadDictionaryFromXML(xmlPoints)

		if not isDead:
			if event.isLoading or event.isRecording:  
				event.GenerateWaveform()
			else:
				levels_path = event.GetAbsLevelsFile()
				try:
					event.levels_list.fromfile(levels_path)
				except LevelsList.CorruptFileError:
					Globals.debug("Cannot load levels from file", levels_path)
				if not event.levels_list:
					event.GenerateWaveform()
			event._Event__UpdateAudioFadePoints()
			event.CreateFilesource()
	
	#_____________________________________________________________________

#=========================================================================

class OpenProjectError(EnvironmentError):
	"""
	This class will get created when a opening a Project fails.
	It's used for handling errors.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self, errno, info = None):
		"""
		Creates a new instance of OpenProjectError.
		
		Parameters:
			errno -- number indicating the type of error:
					1 = invalid uri passed for the Project file.
					2 = unable to unzip the Project.
					3 = Project created by a different version of Jokosher.
					4 = Project file doesn't exist.
					5 = Loading process faild with traceback
			info -- version of Jokosher that created the Project.
					Will be present only along with error #3.
				OR traceback of thrown exception, which will be present with #5
		"""
		EnvironmentError.__init__(self)
		self.info = info
		self.errno = errno
	
	#_____________________________________________________________________
	
#=========================================================================

class CreateProjectError(Exception):
	"""
	This class will get created when creating a Project fails.
	It's used for handling errors.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self, errno, message=None):
		"""
		Creates a new instance of CreateProjectError.
		
		Parameters:
			errno -- number indicating the type of error:
					1 = unable to create a Project object.
					2 = path for Project file already exists.
					3 = unable to create file. (Invalid permissions, read-only, or the disk is full).
					4 = invalid path, name or author.
					5 = invalid uri passed for the Project file.
					6 = unable to load a particular gstreamer plugin (message will be the plugin's name)
			message -- a string with more specific information about the error
		"""
		Exception.__init__(self)
		self.errno = errno
		self.message = message
		
	#_____________________________________________________________________

#=========================================================================

class InvalidProjectError(Exception):
	"""
	This class will get created when there's an invalid Project.
	It's used for handling errors.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self, missingfiles, missingimages):
		"""
		Creates a new instance of InvalidProjectError.
		
		Parameters:
			missingfiles -- filenames of the missing files.
			missingimages -- filenames of the missing images.
		"""
		Exception.__init__(self)
		self.files=missingfiles
		self.images=missingimages
		
	#_____________________________________________________________________

#=========================================================================

class ProjectExportException(Exception):
	"""
	This class will get created when there are problems with the soundcard inputs.
	It's used for handling errors.

	Error Codes:
	MISSING_ELEMENT - incorrect element name or element not installed.
	INVALID_ENCODE_BIN - invalid bin description; invalid syntax, invalid properties, incompatible caps, etc.
	"""
	MISSING_ELEMENT, INVALID_ENCODE_BIN = range(2)
	
	#_____________________________________________________________________
	
	def __init__(self, errno, message):
		"""
		Parameters:
			errno -- number indicating the type of error. See error codes above.
		"""
		Exception.__init__(self)
		self.errno = errno
		self.message = message
		
	#_____________________________________________________________________

#=========================================================================

def ApplyUndoCompat(objectString, functionString, version):
	if UNDO_COMPAT_DICT.has_key(version):
		compact_dict = UNDO_COMPAT_DICT[version]
		tuple_ = (objectString[0], functionString)
		if compact_dict.has_key(tuple_):
			return compact_dict[tuple_]
	
	return functionString

#=========================================================================

JOKOSHER_VERSION_FUNCTIONS = {
	"0.1" : _LoadZPOFile, 
	"0.2" : _LoadZPTFile, 
	"0.9" : _LoadZPNFile,
	"1.0" : _LoadZPNFile,  # 1.0 was never used in a release, and it identical to 0.9
	"0.10" : _LoadZPTenFile,
	"0.11" : _LoadZPTenFile,	# 0.11 is identical to 0.10, exception for project notes, whose presence can be detected
	"0.11.1" : _LoadZPTenFile,
}

zero_nine_compat = {("E", "Move") : "_Compat09_Move"}
zero_two_compat = {
	("E", "Split") : "_Compat02_Split",
	("E", "Join") : "_Compat02_Join",
	("E", "UndoTrim") : "_Compat02_UndoTrim",
}
#we can't import undo from 0.1 because the storage of undo was revamped for version 0.2

#all the compat info from newer versions, also applies to older versions
for key, value in zero_nine_compat.iteritems():
	zero_two_compat.setdefault(key, value)

UNDO_COMPAT_DICT = {"0.2" : zero_two_compat, "0.9" : zero_nine_compat}

#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	AudioPreview.py
#	
#	This module handles previewing audio files from within
#	the gtk open file dialog, when you are browsing for an audio
#	file to import.
#
#-------------------------------------------------------------------------------

import gtk
import gst

#=========================================================================

class AudioPreview(gtk.ToggleButton):
	"""
	A button, added to the load-a-clip-from-a-file open dialog,
	which previews the selected sound.
	"""

	def __init__(self):
		"""
		Creates a new instance of AudioPreview.
		"""
		gtk.ToggleButton.__init__(self, "gtk-media-play")
		self.set_use_stock(True)
		self.uri = None
		self.connect("toggled", self.OnToggle)
		self.connect("destroy", self.OnDestroy)
		# playbin is a gst element which plays a sound and works out everything
		self.previewbin = gst.element_factory_make ("playbin", "preview")
		self.bus = self.previewbin.get_bus()
		self.bus.add_signal_watch()
		# deactivate thyself if the sound finishes or there's an error
		self.bus.connect("message::eos", self.OnEOS)
		self.bus.connect("message::error", self.OnEOS)
		
	#_____________________________________________________________________

	def OnSelection(self, widget):
		"""
		This method is called when a user selects a new file in the import audio file dialog. 
		The preview widget then retrieves the location of the file that has been selected
		and sets the variable self.uri to be the location of the file it retrieves.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.uri = widget.get_preview_uri()
		self.set_active(False)
		
	#_____________________________________________________________________

	def OnToggle(self, widget):
		"""
		This method is called when a user toggles the preview button between a stop
		or playing state. It will then set the button to appear pressed in when playing
		or normal when stopped.
		
		Parameters:
			widget -- reserved for GTK callbacks, dont't use it explicitly.
		"""
		if self.get_active():
			self.previewbin.set_property("uri", self.uri)
			self.previewbin.set_state(gst.STATE_PLAYING)
		else:
			self.previewbin.set_state(gst.STATE_READY)
			
	#_____________________________________________________________________
			
	def OnEOS(self, bus, message):
		"""
		Called when the preview stream ends. It deactives the AudioPreview.
		
		Parameters:
			bus -- reserved for GStreamer callbacks, dont't use it explicitly.
			message -- reserved for GStreamer callbacks, dont't use it explicitly.
		"""
		self.set_active(False)
		
	#_____________________________________________________________________

	def OnDestroy(self, widget=None):
		"""
		This method sets the preview pipeline to gst.STATE_NULL.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.previewbin.set_state(gst.STATE_NULL)
	
	#_____________________________________________________________________

#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Instrument.py
#	
#	This module is the non-gui class the represents instruments. instruments
#	represent a track of audio that can contain many different sources in sequence.
#	This module handles loading and saving instruments from xml, the gstreamer
#	bits for playing and recording events, audio effects plugins, as well as any 
#	instrument specific functionality like; solo, mute, volume, etc.
#
#-------------------------------------------------------------------------------

import pygst
pygst.require("0.10")
import gst
import PlatformUtils
import os, time, shutil
import urlparse # To split up URI's
import gobject
import Event
import UndoSystem, IncrementalSave
import Utils

import Globals
import gettext
_ = gettext.gettext

#=========================================================================	

class Instrument(gobject.GObject):
	"""
	This module is the non-gui class the represents Instruments. Instruments
	represent a track of audio that can contain many different sources in sequence.
	It also handles loading and saving Instruments from xml, the gstreamer
	bits for playing and recording events, audio effects plugins, as well as any 
	Instrument specific functionality like; solo, mute, volume, etc.
	"""
	
	#_____________________________________________________________________
	
	"""
	Only elements that accept these caps will be work if added to the effects bin.
	This current must be audio/x-raw-float to prevent static from occuring
	when adding LADSPA effects while playing.
	"""
	LADSPA_ELEMENT_CAPS = "audio/x-raw-float, width=(int)32, rate=(int)[ 1, 2147483647 ], channels=(int)1, endianness=(int)1234"
	
	"""
	Signals:
		"arm" -- This instrument has been armed or dis-armed for recording.
		"effect" -- The effects for this instrument have changed. See below:
			"effect::added" -- An effect was added to this instrument.
			"effect::removed" -- An effect was removed from this instrument.
			"effect::reordered" -- An effect on this instrument has changed its position.
		"event" -- The events for this instrument have changed. The event ID will be passed as a parameter. See below:
			"event::added" -- An event was added to this instrument.
			"event::removed" -- An event was removed from this instrument.
		"image" -- The image for this instrument has changed.
		"mute" -- This instrument has been muted or unmuted.
		"name" -- The name of this instrument has changed.
		"recording-done" -- The event recording on this instrument has finished recording.
		"selected" -- This instrument has been selected or deselected.
		"visible" -- This instrument has been maximised or minimized in the mix view.
		"volume" -- The volume value for this instrument has changed.
	"""
	
	__gsignals__ = {
		"arm"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"effect"		: ( gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_DETAILED, gobject.TYPE_NONE, () ),
		"event"		: ( gobject.SIGNAL_RUN_LAST | gobject.SIGNAL_DETAILED, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,) ),
		"image"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"mute"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"name"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"recording-done"	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"selected"	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"solo"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"visible"	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () ),
		"volume"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () )
	}
	
	def __init__(self, project, name, type, pixbuf, id=None):
		"""
		Creates a new instance of Instrument.
		
		Parameters:
			project -- the currently active Project.
			name -- name of the Instrument.
			type -- type of the Instrument.
			pixbuf -- image of the Instrument resource object.
			id -- unique ID value for the Instrument.
		"""
		gobject.GObject.__init__(self)
		
		self.project = project
		
		self.events = []				# List of events attached to this instrument
		self.graveyard = []			# List of events that have been deleted (kept for undo)
		self.effects = []				# List of GStreamer effect elements
		
		self.name = name			# Name of this instrument
		self.pixbuf = pixbuf			# The icon pixbuf resource
		self.instrType = type		# The type of instrument
		
		self.isArmed = False			# True if the instrument is armed for recording
		self.isMuted = False			# True if the "mute" button is toggled on
		self.actuallyIsMuted = False	# True if the instrument is muted (silent)
		self.isSolo = False			# True if the instrument is solo'd (only instrument active)
		self.isVisible = True			# True if the instrument should be displayed in the mixer views
		self.isSelected = False		# True if the instrument is currently selected
		
		self.level = 0.0				# Current audio level in range 0..1
		self.volume = 1.0			# Gain of the current instrument in range 0..1
		self.pan = 0.0				# pan number (between -100 and 100)
		self.currentchainpreset = None	# current instrument wide chain preset
		self.output = ""
		self.recordingbin = None
		self.id = project.GenerateUniqueID(id)	#check is id is already being used before setting
		
		self.input = None	# the device to use for recording on this instrument.
		self.inTrack = -1	# Input track to record from if device is multichannel.
	
		# CREATE GSTREAMER ELEMENTS #
		self.playbackbin = gst.element_factory_make("bin", "Instrument_%d"%self.id)
		self.volumeElement = gst.element_factory_make("volume", "Instrument_Volume_%d"%self.id)
		self.levelElement = gst.element_factory_make("level", "Instrument_Level_%d"%self.id)
		self.panElement = gst.element_factory_make("audiopanorama", "Instrument_Pan_%d"%self.id)
		self.resample = gst.element_factory_make("audioresample")
		
		self.composition = gst.element_factory_make("gnlcomposition")
		self.silentGnlSource = gst.element_factory_make("gnlsource")		# the default source that makes the silence between the tracks
		self.silenceAudioSource = gst.element_factory_make("audiotestsrc")
		
		self.effectsBin = gst.element_factory_make("bin", "InstrumentEffects_%d"%self.id)
		self.effectsBinConvert = gst.element_factory_make("audioconvert", "Start_Effects_Converter_%d"%self.id)
		self.effectsBinCaps = gst.element_factory_make("capsfilter", "Effects_float_caps_%d"%self.id)
		self.effectsBinCaps.set_property("caps", gst.Caps(self.LADSPA_ELEMENT_CAPS))
		self.effectsBinEndConvert = gst.element_factory_make("audioconvert", "End_Effects_Converter_%d"%self.id)
		
		self.volumeFadeBin = gst.element_factory_make("bin", "Volume_fades_bin")
		self.volumeFadeElement = gst.element_factory_make("volume", "Volume_Fade_Element")
		self.volumeFadeStartConvert = gst.element_factory_make("audioconvert", "Start_fadebin_converter")	
		self.volumeFadeEndConvert = gst.element_factory_make("audioconvert", "End_fadebin_converter")
		self.volumeFadeOperation = gst.element_factory_make("gnloperation", "gnloperation")
		self.volumeFadeController = gst.Controller(self.volumeFadeElement, "volume")
		
		# CREATE GHOSTPADS FOR BINS #
		self.effectsBin.add(self.effectsBinConvert, self.effectsBinCaps, self.effectsBinEndConvert)
		self.effectsBinSink = gst.GhostPad("sink", self.effectsBinConvert.get_pad("sink"))
		self.effectsBin.add_pad(self.effectsBinSink)
		self.effectsBinSrc = gst.GhostPad("src", self.effectsBinEndConvert.get_pad("src"))
		self.effectsBin.add_pad(self.effectsBinSrc)
		
		self.volumeFadeBin.add(self.volumeFadeElement)
		self.volumeFadeBin.add(self.volumeFadeStartConvert)
		self.volumeFadeBin.add(self.volumeFadeEndConvert)
		volumeFadeBinSink = gst.GhostPad("sink", self.volumeFadeStartConvert.get_pad("sink"))
		self.volumeFadeBin.add_pad(volumeFadeBinSink)
		volumeFadeBinSrc = gst.GhostPad("src", self.volumeFadeEndConvert.get_pad("src"))
		self.volumeFadeBin.add_pad(volumeFadeBinSrc)
		
		# SET ELEMENT PROPERTIES #
		self.levelElement.set_property("interval", gst.SECOND / 50)
		self.levelElement.set_property("message", True)
		self.levelElement.set_property("peak-ttl", 0)
		self.levelElement.set_property("peak-falloff", 20)
		
		self.panElement.set_property("panorama", 0)

		self.silenceAudioSource.set_property("wave", 4)	#4 is silence
		
		self.silentGnlSource.set_property("priority", 2 ** 32 - 1)
		self.silentGnlSource.set_property("start", 0)
		self.silentGnlSource.set_property("duration", 1000 * gst.SECOND)
		self.silentGnlSource.set_property("media-start", 0)
		self.silentGnlSource.set_property("media-duration", 1000 * gst.SECOND)
		
		self.volumeFadeOperation.set_property("start", long(0) * gst.SECOND)
		self.volumeFadeOperation.set_property("duration", long(20) * gst.SECOND)
		self.volumeFadeOperation.set_property("priority", 1)
		
		self.volumeFadeController.set_interpolation_mode("volume", gst.INTERPOLATE_LINEAR)
		
		# ADD ELEMENTS TO THE PIPELINE AND/OR THEIR BINS #
		self.playbackbin.add(self.volumeElement, self.levelElement, self.panElement, self.resample)
		self.playbackbin.add(self.composition)
		self.playbackbin.add(self.effectsBin)
		
		self.volumeFadeOperation.add(self.volumeFadeBin)
		self.silentGnlSource.add(self.silenceAudioSource)
		self.composition.add(self.silentGnlSource)
		self.composition.add(self.volumeFadeOperation)
		
		# LINK GSTREAMER ELEMENTS #
		self.effectsBinConvert.link(self.effectsBinCaps)
		self.effectsBinCaps.link(self.effectsBinEndConvert)
		
		self.effectsBin.link(self.volumeElement)
		self.volumeElement.link(self.levelElement)
		self.levelElement.link(self.panElement)	
		self.panElement.link(self.resample)

		self.playghostpad = gst.GhostPad("src", self.resample.get_pad("src"))
		self.playbackbin.add_pad(self.playghostpad)
		
		self.volumeFadeStartConvert.link(self.volumeFadeElement)
		self.volumeFadeElement.link(self.volumeFadeEndConvert)
		
		self.AddAndLinkPlaybackbin()

		self.composition.connect("pad-added", self.__PadAddedCb)
		self.composition.connect("pad-removed", self.__PadRemovedCb)
		
		#mute this instrument if another one is solo
		self.OnMute()
		#set the volume element since it depends on the project's volume as well
		self.UpdateVolume()
		
	#_____________________________________________________________________
	
	def __PadAddedCb(self, element, pad):
		"""
		Links a new pad to the rest of the playbackbin when one is created
		by the composition.
		
		Parameters:
			element -- GStreamer element calling this function.
			pad -- newly added pad object.
		"""
		Globals.debug("NEW PAD on instrument %s" % self.name)
		convpad = self.effectsBin.get_compatible_pad(pad, pad.get_caps())
		pad.link(convpad)

	#_____________________________________________________________________

	def __PadRemovedCb(self, element, pad):
		"""
		Removes a GStreamer pad from the specified instrument.
		
		Parameters:
			element -- GStreamer element calling this function.
			pad -- pad to be removed from the Instrument.
		"""
		Globals.debug("pad removed on instrument %s" % self.name)
		self.composition.set_state(gst.STATE_READY)

	#_____________________________________________________________________
	
	def __repr__(self):
		"""
		Creates a representation string of the Instrument.

		Returns:
			a string representing the id and name for this Instrument.
		"""
		return "Instrument [%d] %s"%(self.id, self.name)
		
	#_____________________________________________________________________
		
	def StoreToXML(self, doc, parent, graveyard=False):
		"""
		Converts this Instrument into an XML representation suitable for saving to a file.

		Parameters:
			doc -- the XML document object the Instrument will be saved to.
			parent -- the parent node that the serialized Instrument should
						be added to.
			graveyard -- True if this Instrument is on the graveyard stack,
						and should be serialized as a dead Instrument.
		"""
		if graveyard:
			ins = doc.createElement("DeadInstrument")
		else:
			ins = doc.createElement("Instrument")
		parent.appendChild(ins)
		ins.setAttribute("id", str(self.id))
		
		items = ["name", "isArmed", 
				"isMuted", "isSolo", "input", "output", "volume",
				"isSelected", "isVisible", "inTrack", "instrType", "pan"]
		
		params = doc.createElement("Parameters")
		ins.appendChild(params)
		
		Utils.StoreParametersToXML(self, doc, params, items)
		
		for effect in self.effects:
			globaleffect = doc.createElement("GlobalEffect")
			globaleffect.setAttribute("element", effect.get_factory().get_name())
			ins.appendChild(globaleffect)
		
			propsdict = {}
			for prop in gobject.list_properties(effect):
				if prop.flags & gobject.PARAM_WRITABLE:
					propsdict[prop.name] = effect.get_property(prop.name)
			
			Utils.StoreDictionaryToXML(doc, globaleffect, propsdict)
			
		for ev in self.events:
			ev.StoreToXML(doc, ins)
		for ev in self.graveyard:
			ev.StoreToXML(doc, ins, graveyard=True)

	#_____________________________________________________________________

	def AddEffect(self, effectName):
		"""
		Adds an effect to the pipeline for this Instrument.
		
		Considerations:
			The effect is always placed in the pipeline after any other
			effects that were previously added.
			
		Parameters:
			effectName -- GStreamer element name of the effect to add.
			
		Returns:
			the added effect element.
		"""
		#make the new effect
		effectElement = gst.element_factory_make(effectName)
		self.effects.append(effectElement)
		#add the element to effects bin
		self.effectsBin.add(effectElement)
		
		# The src pad on the first element (an audioconvert) in the bin
		startSrcPad = self.effectsBinConvert.get_pad("src")
		
		state = self.playbackbin.get_state(0)[1]
		if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING:
			startSrcPad.set_blocked(True)
			
		lastEffectElement = self.effectsBinEndConvert.get_pad("sink").get_peer().get_parent()
		
		lastEffectElement.unlink(self.effectsBinEndConvert)
		lastEffectElement.link(effectElement)
		effectElement.link(self.effectsBinEndConvert)
		
		# make the element's state match the bin's state
		effectElement.set_state(state)
		#give it a lambda for a callback that does nothing, so we don't have to wait
		startSrcPad.set_blocked_async(False, lambda x,y: False)
		
		self.emit("effect::added")
		
		return effectElement
	
	#_____________________________________________________________________
	
	def RemoveEffect(self, effect):
		"""
		Remove the given GStreamer element from the effects bin.
		
		Parameters:
			effect -- GStreamer effect to be removed from this Instrument.
		"""
		if effect not in self.effects:
			Globals.debug("Error: trying to remove an element that is not in the list")
			return
		
		for pad in effect.sink_pads():
			if pad.is_linked():
				previousElement = pad.get_peer().get_parent()
				break
		
		for pad in effect.src_pads():
			if pad.is_linked():
				nextElement = pad.get_peer().get_parent()
				break
		
		# The src pad on the first element (an audioconvert) in the bin
		startSrcPad = self.effectsBinConvert.get_pad("src")
		
		state = self.playbackbin.get_state(0)[1]
		if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING:
			startSrcPad.set_blocked(True)
		
		previousElement.unlink(effect)
		effect.unlink(nextElement)
		previousElement.link(nextElement)
		
		# Remove and dispose of the element
		self.effectsBin.remove(effect)
		effect.set_state(gst.STATE_NULL)
		#remove the effect from our own list
		self.effects.remove(effect)
		
		#give it a lambda for a callback that does nothing, so we don't have to wait
		startSrcPad.set_blocked_async(False, lambda x,y: False)
		
		self.emit("effect::removed")
	
	#_____________________________________________________________________
	
	def ChangeEffectOrder(self, effect, newPosition):
		"""
		Move a given GStreamer element inside the effects bin. This method
		does not swap the element into its new position, it simply shifts all the
		elements between the effect's current position and the new position
		down by one. Since a Gstreamer pipeline is essentially a linked list, this
		"shift" implementation is the fastest way to make one element move to
		a new position without changing the order of any of the others.
		For example if this instrument has five effects which are ordered
		A, B, C, D, E, and I call this method with effect D and a new position
		of 1 the new order will be:  A, D, B, C, E.
		
		Parameters:
			effect -- GStreamer effect to be moved.
			newPosition -- value of the new position inside the effects bin
					the effect will have (with 0 as the first position).
		"""
		if effect not in self.effects:
			Globals.debug("Error: trying to remove an element that is not in the list")
			return
		if newPosition >= len(self.effects):
			Globals.debug("Error: trying to move effect to position past the end of the list")
			return
			
		oldPosition = self.effects.index(effect)
		if oldPosition == newPosition:
			#the effect is already in the proper position
			return
		
		# The effect currently at the position we want to move the given effect to
		newPositionEffect = self.effects[newPosition]
		
		for pad in effect.sink_pads():
			if pad.is_linked():
				previousElement = pad.get_peer().get_parent()
				break
		
		for pad in effect.src_pads():
			if pad.is_linked():
				nextElement = pad.get_peer().get_parent()
				break
		
		# The src pad on the first element (an audioconvert) in the bin
		startSrcPad = self.effectsBinConvert.get_pad("src")
		# check the state and block if we have to
		state = self.playbackbin.get_state(0)[1]
		if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING:
			startSrcPad.set_blocked(True)
			
		#here's where we start unlinking and relinking
		previousElement.unlink(effect)
		effect.unlink(nextElement)
		previousElement.link(nextElement)
		
		if oldPosition > newPosition:
			for pad in newPositionEffect.sink_pads():
				if pad.is_linked():
					newPositionPrevious = pad.get_peer().get_parent()
					break
			
			newPositionPrevious.unlink(newPositionEffect)
			newPositionPrevious.link(effect)
			effect.link(newPositionEffect)
		else:
			for pad in newPositionEffect.src_pads():
				if pad.is_linked():
					newPositionNext = pad.get_peer().get_parent()
					break
			
			newPositionEffect.unlink(newPositionNext)
			newPositionEffect.link(effect)
			effect.link(newPositionNext)
			
		# remove and insert to our own llst so it matches the changes just made
		del self.effects[oldPosition]
		self.effects.insert(newPosition, effect)
		
		#give it a lambda for a callback that does nothing, so we don't have to wait
		startSrcPad.set_blocked_async(False, lambda x,y: False)
		
		self.emit("effect:reordered")
	
	#_____________________________________________________________________
	
	def GetRecordingEvent(self):
		"""
		Obtain an Event suitable for recording.
		Returns:
			an Event suitable for recording.
		"""
		event = Event.Event(self)
		event.start = self.project.transport.GetPosition()
		event.isRecording = True
		event.name = _("Recorded audio")
		
		ext = Globals.settings.recording["file_extension"]
		filename = "%s_%d.%s" % (Globals.FAT32SafeFilename(self.name), event.id, ext)
		event.file = filename
		event.levels_file = filename + Event.Event.LEVELS_FILE_EXTENSION
		
		inc = IncrementalSave.NewEvent(self.id, filename, event.start, event.id, recording=True)
		self.project.SaveIncrementalAction(inc)
		
		#must add it to the instrument's list so that an update of the event lane will not remove the widget
		self.events.append(event)
		self.emit("event::added", event)
		return event

	#_____________________________________________________________________

	@UndoSystem.UndoCommand("DeleteEvent", "temp", incremental_save=False)
	def FinishRecordingEvent(self, event):
		"""
		Called to log the adding of this event on the undo stack
		and to properly load the file that has just been recorded.
		
		Parameters:
			event -- Event object that has finished being recorded.
		"""
		event.isRecording = False
		event.GenerateWaveform()
		self.temp = event.id
		self.emit("recording-done")
	
	#_____________________________________________________________________
	
	def FinalizeRecording(self, event):
		"""
		Called when the recording of an Event has finished.
		
		Parameters:
			event -- Event object that has finished being recorded.
		"""
		#create our undo action to make everything atomic
		undoAction = self.project.NewAtomicUndoAction()
		#make sure the event will act mormally (like a loaded file) now
		self.FinishRecordingEvent(event, _undoAction_=undoAction)
		# remove all the events behind the recorded event (because we can't have overlapping events.
		self.RemoveEventsUnderEvent(event, undoAction)
		
	#_____________________________________________________________________
	
	def AddEventsFromList(self, start, fileList, copyFile=False, undoAction=None):
		"""
		Adds multiple files to an instrument one after another starting
		at the given start position.
		
		Parameters:
			start -- the offset time in seconds for the first event.
			fileList -- URIs to the Event files.
			copyfile --	True = copy the files to Project's audio directory.
						False = don't copy the files to the Project's audio directory.
		"""
		if not fileList:
			return
			
		if not undoAction:
			undoAction = self.project.NewAtomicUndoAction()
			
		for uri in fileList:
			# Parse the uri, and continue only if it is pointing to a local file
			(scheme, domain, file, params, query, fragment) = urlparse.urlparse(uri, "file", False)
			if scheme == "file":
				file = PlatformUtils.url2pathname(file)
				event = self.addEventFromFile(start, file, copyFile, _undoAction_=undoAction)
			else:
				event = self.addEventFromURL(start, uri, _undoAction_=undoAction)
			
			if event:
				event.MoveButDoNotOverlap(event.start)
				event.SetProperties()
				start += event.duration
	
	#_____________________________________________________________________

	@UndoSystem.UndoCommand("DeleteEvent", "temp", incremental_save=False)
	def addEventFromFile(self, start, file, copyfile=False, name=None, duration=None, levels_file=None):
		"""
		Adds an Event from a file to this Instrument.
		
		Parameters:
			start -- the offset time in seconds for the Event.
			file -- path to the Event file.
			copyfile --	True = copy the file to Project's audio directory.
						False = don't copy the file to the Project's audio directory.
			name -- An optional user visible name. The filename will be used if None.
						
		Returns:
			the added Event.
		"""
		filelabel=file
		event_id = self.project.GenerateUniqueID(None,  reserve=False)
		if not name:
			name = os.path.basename(file)
		root,  extension = os.path.splitext(name.replace(" ", "_"))
		
		if extension:
			newfile = "%s_%d%s" % (root, event_id, extension)
		else:
			newfile = "%s_%d" % (root, event_id)

		if copyfile:
			audio_file = os.path.join(self.project.audio_path, newfile)
			
			try:
				shutil.copyfile(file, audio_file)
			except IOError:
				raise UndoSystem.CancelUndoCommand()
				
			self.project.deleteOnCloseAudioFiles.append(audio_file)
			inc = IncrementalSave.NewEvent(self.id, newfile, start, event_id)
			self.project.SaveIncrementalAction(inc)
			
			file = newfile
		else:
			inc = IncrementalSave.NewEvent(self.id, file, start, event_id)
			self.project.SaveIncrementalAction(inc)

		ev = Event.Event(self, file, event_id, filelabel)
		ev.start = start
		ev.name = name
		self.events.append(ev)
		
		if duration and levels_file:
			ev.duration = duration
			ev.levels_file = levels_file
			ev.levels_list.fromfile(ev.GetAbsLevelsFile())
			# update properties and position when duration changes.
			ev.MoveButDoNotOverlap(ev.start)
			ev.SetProperties()
		else:
			ev.GenerateWaveform()

		self.temp = ev.id
		
		self.emit("event::added", ev)
		
		return ev
		
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("DeleteEvent", "temp", incremental_save=False)
	def addEventFromURL(self, start, url):
		"""
		Adds an Event from a URL to this Instrument.
		
		Considerations:
			Unlike addEventFromFile, there is no copyfile option here,
			because it's mandatory.
		
		Parameters:
			start -- The offset time in seconds for the Event.
			url -- url of the Event to be added.
			
		Returns:
			the added Event.
		"""
		event_id = self.project.GenerateUniqueID(None,  reserve=False)
		# no way of knowing whether there's a filename, so make one up
		newfile = str(event_id)
		
		audio_file = os.path.join(self.project.audio_path, newfile)
		self.project.deleteOnCloseAudioFiles.append(audio_file)
		
		# Create the event now so we can return it, and fill in the file later
		ev = Event.Event(self, newfile, event_id, url)
		ev.start = start
		ev.name = os.path.split(audio_file)[1]
		ev.isDownloading = True
		self.events.append(ev)
		
		Globals.debug("Event data downloading...")
		result = ev.CopyAndGenerateWaveform(url)
		
		if not result:
			self.events.remove(ev)
			raise UndoSystem.CancelUndoCommand()
		
		inc = IncrementalSave.StartDownload(self.id, url, newfile, start, event_id)
		self.project.SaveIncrementalAction(inc)
		
		self.temp = ev.id
		self.emit("event::added", ev)
		
		return ev
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("DeleteEvent", "temp")
	def addEventFromEvent(self, start, event):
		"""
		Creates a new Event instance identical to the given Event object
		and adds it to this instrument (for paste functionality).
		
		Parameters:
			start - the offset time in seconds for the new Event.
			event - the Event to be cloned on this instrument.
		"""
		ev = Event.Event(self, event.file)
		ev.start = start
		for i in ["duration", "name", "offset"]:
			setattr(ev, i, getattr(event, i))
		ev.levels_list = event.levels_list.copy()
		ev._Event__fadePointsDict = event._Event__fadePointsDict.copy()
		ev._Event__UpdateAudioFadePoints()
		
		self.events.append(ev)
		ev.SetProperties()
		ev.MoveButDoNotOverlap(ev.start)
		
		self.temp = ev.id
		self.emit("event::added", ev)
		return ev
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("ResurrectEvent", "temp")
	def DeleteEvent(self, eventid):
		"""
		Removes an Event from this Instrument.
			
		Parameters:
			eventid -- ID of the Event to be removed.
		"""
		event = [x for x in self.events if x.id == eventid][0]
		
		self.graveyard.append(event)
		self.events.remove(event)
		event.DestroyFilesource()
		event.StopGenerateWaveform(False)
		
		self.temp = eventid
		self.emit("event::removed", event)
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("DeleteEvent", "temp")
	def ResurrectEvent(self, eventid):
		"""
		Brings an Event back from the graveyard.
		
		Parameters:
			eventid -- ID of the Event to be resurrected.
		"""
		event = [x for x in self.graveyard if x.id == eventid][0]
		
		self.events.append(event)
		self.graveyard.remove(event)
		event.CreateFilesource()
		if event.isLoading or not event.levels_list:
			event.GenerateWaveform()
		
		self.temp = eventid
		self.emit("event::added", event)
	
	#_____________________________________________________________________

	def MultipleEventsSelected(self):
		"""
		Confirms whether or not multiple events are selected.
		
		Returns:
			True = multiple Instruments have been selected.
			False = none or just one Instrument has been seleced.
		"""
		multiple = 0
		for ev in self.events:
			if (ev.isSelected):
				multiple += 1
		return (multiple > 1)

	#_____________________________________________________________________

	def SetLevel(self, level):
		"""
		Sets the level of this Instrument.
		
		Considerations:
			This sets the current REPORTED level, NOT THE VOLUME!
		
		Parameters:
			level -- new level value in a [0,1] range.
		"""
		self.level = level
	
	#_____________________________________________________________________

	def SetVolume(self, volume):
		"""
		Sets the volume of this Instrument.
		
		Parameters:
			volume -- new volume value in a [0,1] range.
		"""
		if self.volume != volume:
			self.volume = volume
			self.UpdateVolume()
			self.emit("volume")

	#_____________________________________________________________________
	
	def CommitVolume(self):
		"""
		Signal that the volume is no longer volatile. This means we can incrementally save,
		which we didn't do before because the volume was rapidly changing as the user
		moved the mouse.
		"""
		inc = IncrementalSave.InstrumentSetVolume(self.id, self.volume)
		self.project.SaveIncrementalAction(inc)
		
	#_____________________________________________________________________
	
	def UpdateVolume(self):
		"""
		Updates the volume property of the gstreamer volume element
		based on this instrument's volume and the project's master volume.
		"""
		volume = self.volume * self.project.volume
		self.volumeElement.set_property("volume", volume)
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("SetName", "temp")
	def SetName(self, name):
		"""
		Sets the Instrument's name so it can be registered in the undo stack.
		
		Parameters:
			name -- string with the name for the Instrument.
		"""
		if self.name != name:
			self.temp = self.name
			self.name = name
			self.emit("name")
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("ToggleArmed")
	def ToggleArmed(self):
		"""
		Arms/Disarms the Instrument for recording.
		"""
		self.isArmed = not self.isArmed
		self.emit("arm")
		
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("ToggleMuted", "temp")
	def ToggleMuted(self, wasSolo):
		"""
		Mutes/Unmutes the Instrument.
		
		Parameters:
			wasSolo --	True = the Instrument had solo mode enabled.
						False = the Instrument was not in solo mode.
						
		Considerations:
			The signal "mute" is not emitted here because it is emitted in
			the OnMute() function.
		"""
		self.temp = self.isSolo
		self.isMuted = not self.isMuted
		if self.isSolo and not wasSolo:
			self.isSolo = False
			self.project.soloInstrCount -= 1
			self.project.OnAllInstrumentsMute()
		elif not self.isSolo and wasSolo:
			self.isSolo = True
			self.project.soloInstrCount += 1
			self.project.OnAllInstrumentsMute()
		else:
			self.OnMute()
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("ToggleSolo", "temp")
	def ToggleSolo(self, wasMuted):
		"""
		Mutes/Unmutes the other Instruments in the Project.
		
		Parameters:
			wasMuted --	True = the Instrument was muted.
						False = the Instrument was not muted.
		"""
		self.temp = self.isMuted
		self.isMuted = wasMuted
		
		if self.isSolo:
			self.isSolo = False
			self.project.soloInstrCount -= 1
		else:
			self.isSolo = True
			self.project.soloInstrCount += 1
		
		self.project.OnAllInstrumentsMute()
		self.emit("solo")
	
	#_____________________________________________________________________
	
	@UndoSystem.UndoCommand("SetVisible", "temp")
	def SetVisible(self, visible):
		"""
		Sets whether the Instrument is minimized in the CompactMixView.
		
		Parameters:
			visible --	True = the Instrument should be hidden in the mixing view.
						False = the Instrument should be shown normally in 
								the mixing view.
		"""
		if self.isVisible != visible:
			self.temp = self.isVisible
			self.isVisible = visible
			self.emit("visible")
	
	#_____________________________________________________________________
	
	def SetSelected(self, sel):
		"""
		Sets the Instrument to be highlighted and receive keyboard actions.
		
		Parameters:
			sel -- 	True = the Instrument is currently selected.
					False = the Instrument is not currently selected.
		"""
		# No need to emit signal when there is no change in selection state
		if self.isSelected is not sel:
			self.isSelected = sel
			self.emit("selected")
	
	#_____________________________________________________________________
	
	def SetInput(self, device, inTrack):
		if device != self.input or inTrack != self.inTrack:
			self.input = device
			self.inTrack = inTrack
			
			inc = IncrementalSave.InstrumentSetInput(self.id, device, inTrack)
			self.project.SaveIncrementalAction(inc)
			
	#_____________________________________________________________________
	
	def OnMute(self):
		"""
		Updates the GStreamer volume element to reflect the mute status.
		"""
		self.checkActuallyIsMuted()
		if self.actuallyIsMuted:
			self.volumeElement.set_property("mute", True)
		else:
			self.volumeElement.set_property("mute", False)
		
		self.emit("mute")
	
	#_____________________________________________________________________
	
	def checkActuallyIsMuted(self):
		"""
		Determines if this Intrument should be muted, by taking into account
		if any other Intruments are muted.
		"""
		if self.isMuted:
			self.actuallyIsMuted = True
		elif self.isSolo:
			self.actuallyIsMuted = False
		elif self.project.soloInstrCount > 0:
			self.actuallyIsMuted = True
		else:
			self.actuallyIsMuted = False
	
	#_____________________________________________________________________
	
	def AddAndLinkPlaybackbin(self):
		"""
		Creates a playback bin for this Instrument and adds it to the main
		playback pipeline. *CHECK*
		"""
		#make sure our playbackbin is in the same state so the pipeline can continue what it was doing
		status, state, pending = self.project.playbackbin.get_state(0)
		if pending != gst.STATE_VOID_PENDING:
			self.playbackbin.set_state(pending)
		else:
			self.playbackbin.set_state(state)
			
		if not self.playbackbin in list(self.project.playbackbin.elements()):
			self.project.playbackbin.add(self.playbackbin)
			Globals.debug("added instrument playbackbin to adder playbackbin", self.id)
		if not self.playghostpad.get_peer():
			self.playbackbin.link(self.project.adder)
			#give it a lambda for a callback that does nothing, so we don't have to wait
			self.playghostpad.set_blocked_async(False, lambda x,y: False)
			Globals.debug("linked instrument playbackbin to adder (project)")

	#_____________________________________________________________________
	
	def RemoveAndUnlinkPlaybackbin(self):
		"""
		Removes this Instrumen's playback bin from the main playback pipeline. *CHECK*
		"""
		#get reference to pad before removing self.playbackbin from project.playbackbin!
		pad = self.playghostpad.get_peer()
		
		if pad:
			status, state, pending = self.playbackbin.get_state(0)
			if state == gst.STATE_PAUSED or state == gst.STATE_PLAYING or \
					pending == gst.STATE_PAUSED or pending == gst.STATE_PLAYING:
				self.playghostpad.set_blocked(True)
			self.playbackbin.unlink(self.project.adder)
			self.project.adder.release_request_pad(pad)
			Globals.debug("unlinked instrument playbackbin from adder")
		
		if self.playbackbin in list(self.project.playbackbin.elements()):
			self.project.playbackbin.remove(self.playbackbin)
			Globals.debug("removed instrument playbackbin from project playbackbin")
	
	#_____________________________________________________________________

	@UndoSystem.UndoCommand("ChangeType", "temp", "temp2")
	def ChangeType(self, type, name):
		"""
		Changes the Intrument's type and name.
		
		Considerations:
			The given type must be loaded in the Instrument list
			in Globals or the image will not be found.
			
		Parameters:
			type -- new type for this Instrument.
			name -- new name for this Instrument.
		"""
		
		self.temp = self.instrType
		self.temp2 = self.name
		
		pixbuf = Globals.getCachedInstrumentPixbuf(type)
		if type == self.instrType or not pixbuf:
			raise UndoSystem.CancelUndoCommand()

		for tuple_ in Globals.getCachedInstruments():
			if tuple_[1] == self.instrType:
				if tuple_[0] == self.name:
					#if the instrument name has not been modified by the user, we can replace it.
					self.name = name
					self.emit("name")
				break

		self.instrType = type
		self.pixbuf = pixbuf

		self.emit("image")

	#_____________________________________________________________________

	def PrepareController(self):
		"""
		Fills the gst.Controller for this Instrument with its list of fade times.
		"""
		
		Globals.debug("Preparing the controller")
		# set the length of the operation to be the full length of the project
		self.volumeFadeOperation.set_property("duration", self.project.GetProjectLength() * gst.SECOND)
		self.volumeFadeController.unset_all("volume")
		firstpoint = False
		for ev in self.events:
			if not ev.audioFadePoints:
				#there are no fade points, so just make it 100% all the way through
				for point, vol in ((ev.start, 0.99), (ev.start+ev.duration, 0.99)):
					Globals.debug("FADE POINT: time(%.2f) vol(%.2f)" % (point, vol))
					self.volumeFadeController.set("volume", (point) * gst.SECOND, vol)
				continue
			
			for point in ev.audioFadePoints:
				if ev.start + point[0] == 0.0:
					firstpoint = True
				#FIXME: remove vol=0.99 when gst.Controller is fixed to accept many consecutive 1.0 values.
				if point[1] == 1.0:
					vol = 0.99
				else:
					vol = point[1]
				Globals.debug("FADE POINT: time(%.2f) vol(%.2f)" % (ev.start + point[0], vol))
				self.volumeFadeController.set("volume", (ev.start + point[0]) * gst.SECOND, vol)
		if not firstpoint:
			Globals.debug("Set extra zero fade point")
			self.volumeFadeController.set("volume", 0, 0.99)
	#_____________________________________________________________________
	
	def RemoveEventsUnderEvent(self, mainEvent, undoAction=None):
		"""
		Deletes and/or trims any events which are between the two given values.
		This is useful for ensuring that a particular section of this instrument in
		totally clear of any audio.
		
		Parameters:
			start -- the start of the interval in seconds
			stop -- the stop of the interval in seconds
		"""
		if not undoAction:
			#init our own undo action so everything is nice and atomic
			undoAction = self.project.NewAtomicUndoAction()
		
		start = mainEvent.start
		stop = mainEvent.start + max(mainEvent.duration, mainEvent.loadingLength)
		leftTrimEvent = rightTrimEvent = None
		for event in self.events:
			if event is mainEvent:
				continue
			
			eventLeft = event.start
			eventRight = event.start + event.duration
			if start <= eventLeft and eventRight <= stop:
				#this event is in between
				self.DeleteEvent(event.id, _undoAction_=undoAction)
			elif eventLeft < start < eventRight and eventRight <= stop:
				# this event is straddling the start of our interval
				leftTrimEvent = event
			elif start <= eventLeft and eventLeft < stop < eventRight:
				# this event is straddling the stop of our interval
				rightTrimEvent = event
		
		if leftTrimEvent:
			leftPiece = leftTrimEvent.SplitEvent(start - leftTrimEvent.start, _undoAction_=undoAction)
			self.DeleteEvent(leftPiece.id, _undoAction_=undoAction)
		if rightTrimEvent:
			rightTrimEvent.SplitEvent(stop - rightTrimEvent.start, _undoAction_=undoAction)
			self.DeleteEvent(rightTrimEvent.id, _undoAction_=undoAction)
	
	#_____________________________________________________________________
#=========================================================================
